"""eMews service CORE 9 interface."""
from typing import Any, Dict, List

import pathlib

from core.config import Configuration, ConfigString
from core.services.base import CoreService, ServiceMode

SERVICE_NAME: str = "helloworld"
SERVICE_DISPLAY_NAME: str = "HelloWorld"


class EmewsServiceHelloworld(CoreService):
    """eMews service class template.  CORE 9 Service paradigm."""

    name: str = SERVICE_DISPLAY_NAME
    group: str = "eMews"
    directories: List[str] = []
    files: List[str] = [f"emews_service_{SERVICE_NAME}.sh"]
    executables: List[str] = ["python3"]
    dependencies: List[str] = ["HostNode"]
    startup: List[str] = [f"sh emews_service_{SERVICE_NAME}.sh"]
    validate: List[str] = []
    shutdown: List[str] = []
    validation_mode: ServiceMode = ServiceMode.NON_BLOCKING
    default_configs: List[Configuration] = [
        ConfigString(id="emews_service_config", label="Service Configuration",
                     options=["default", "distworldclient", "distworldserver"], default="default"),
    ]
    modes: Dict[str, Dict[str, str]] = {}

    def data(self) -> Dict[str, Any]:
        emews_pkg_path = pathlib.Path(__file__).parent.parent.parent.parent.resolve()
        return {
            'emews_pkg_path': emews_pkg_path,
            'emews_service_name': SERVICE_NAME,
        }

    def get_text_template(self, name: str) -> str:
        assert name == f"emews_service_{SERVICE_NAME}.sh"
        return """
        #!/bin/sh
        export PYTHONPATH=${emews_pkg_path}
        python3 -m emews -n ${node.name} servspawn ${emews_service_name} ${config['emews_service_config']} 1>> emews_servspawn.log 2>&1
        """



"""Hub process nodes.  These nodes connect to the hub for monitoring, sending commands, etc."""
from typing import Mapping

import logging
import pickle
import sys
import threading

from emews.base.client import ClientNode, SynchronizedClientNode
from emews.base.node import ConsoleFormatter, OutboundEvent, NetProtocolID, NET_PROTOCOL, NODE_ID_HUB, \
    NODE_TYPE_HUBCONSOLE, NODE_TYPE_HUBMON, STR_ENCODING


class HubConsole(SynchronizedClientNode):
    """Provides an interface to the hub for giving commands."""

    __slots__ = ('_console_thread', '_console_event', '_console_counter', '_current_command',
                 '_prompt', '_thread_exit_counter', '_console_exit', '_console_eof')

    def __init__(self, system_config: Mapping):
        """Constructor."""
        super().__init__(system_config, NODE_TYPE_HUBCONSOLE)

        import readline  # importing this module adds command history to input()

        NET_PROTOCOL[NetProtocolID.HUB_HUBCONSOLE_WEL].response_handler = self._response_console_welcome
        NET_PROTOCOL[NetProtocolID.HUB_CONSOLE_COMMAND].response_handler = self._response_console_command

        self._console_thread = threading.Thread(name="console", target=self._console, daemon=True)

        self._console_event = threading.Event()
        self._console_counter = self._add_counter(self._handle_console_counter)
        self._thread_exit_counter = self._add_counter(self._handle_thread_exit_counter)

        self._console_exit = False
        self._console_eof = False

        self._current_command: str = ''
        self._prompt = ":>"

    def start(self) -> None:
        """Start the node."""
        self._send_hub_request(NetProtocolID.HUB_HUBCONSOLE_WEL)
        super().start()

    def shutdown(self) -> None:
        """Handle shutdown tasks post event-loop stop."""
        sys.stdout.write("\n")

    def _shutdown_signal_handler(self, signum: int, frame):
        """Signal handler for incoming signals (those which are registered)."""
        if not self._hub_conn.is_connected:
            super()._shutdown_signal_handler(signum, frame)
            return

        sys.stdout.write(f"Use CTRL-D to exit console\n{self._prompt}")
        sys.stdout.flush()

    def _on_scenario_start(self) -> None:
        pass

    def _on_scenario_end(self) -> None:
        pass

    def _connect_hub(self) -> None:
        """Initiate an outbound connection attempt to the hub node."""
        sys.stdout.write(f"Connecting to the hub node at {self._hub_conn.peer_address}, port {self._hub_port} ...\n")
        sys.stdout.flush()
        super()._connect_hub()

    def _on_peer_disconnection(self, peer_node_id: int, log_msg: str) -> None:
        """Handle hub node disconnection."""
        sys.stdout.write(f"\nDisconnected from hub node: {log_msg}\n")
        sys.stdout.flush()
        super()._on_peer_disconnection(peer_node_id, log_msg)

    def _on_hub_connect_failed(self, hub_address: str, reason: str) -> None:
        """Connection attempt to the hub failed."""
        sys.stdout.write(f"Could not connect to the hub node: {reason}\n")
        sys.stdout.flush()
        super()._on_hub_connect_failed(hub_address, reason)

    def _reconnect_timer(self):
        """Create a new timer to act as a delay for peer reconnection attempt."""
        sys.stdout.write(f"Waiting {self._hub_reconnect_delay} seconds before trying to connect again...\n")
        sys.stdout.flush()
        super()._reconnect_timer()

    def _on_outbound_connected(self, outbound_conn_event: OutboundEvent,
                               received_node_id: int, received_node_type: int) -> None:
        """Connection attempt to the hub node succeeded."""
        super()._on_outbound_connected(outbound_conn_event, received_node_id, received_node_type)
        sys.stdout.write(f"Connected to the hub node\n\n")
        sys.stdout.flush()

    def _console(self):
        """Console thread.  Provides a console for user input."""
        while not self._console_exit:
            try:
                command = input(self._prompt)
            except EOFError:
                # ctrl-D pressed
                self._console_eof = True
                break

            command = command.strip()
            if command == '':
                continue

            self._current_command = command
            self._console_counter.increment()

            self._console_event.wait()  # wait for command to process (handle_console_counter())
            self._console_event.clear()

        self._thread_exit_counter.increment()

    # event handlers
    def _handle_console_counter(self, counter_event, num_increments):
        """Process current command (main thread execution)."""
        assert counter_event is self._console_counter
        assert num_increments == 1

        command = self._current_command

        if command.split()[0] == 'exit':
            # ignore any args that may be given with the command
            self._console_exit = True
            self._console_event.set()
            return

        # send command to hub
        self._send_hub_request(NetProtocolID.HUB_CONSOLE_COMMAND, request_data=(bytes(command, STR_ENCODING),))

    def _handle_thread_exit_counter(self, counter_event, num_increments):
        """Process console thread exit (main thread execution)."""
        assert counter_event is self._thread_exit_counter

        if self._console_eof:
            sys.stdout.write("\n\nConsole session finished (EOF)")
        else:
            sys.stdout.write("\nConsole session finished (exit)")
        sys.stdout.flush()

        self.stop()

    # net responses
    def _response_console_welcome(self, node_id: int, local_data, hub_name: bytes, welcome_msg: bytes):
        """Response from hub node for welcome msg."""
        assert node_id == NODE_ID_HUB

        self._prompt = f"{ConsoleFormatter.STYLE_BOLD}{self.node_name}@{hub_name.decode()}" \
                       f"{self._prompt}{ConsoleFormatter.STYLE_ENDC} "

        sys.stdout.write(f"{welcome_msg.decode()}\n\n")
        sys.stdout.write("Type 'exit' to exit console.\n")
        sys.stdout.flush()
        self._console_thread.start()

    def _response_console_command(self, node_id: int, local_data, cmd_result: bytes):
        """Response from hub to command issued."""
        assert node_id == NODE_ID_HUB

        sys.stdout.write(cmd_result.decode())
        sys.stdout.flush()

        self._console_event.set()


class HubMon(ClientNode):
    """Provides an interface to the hub for monitoring."""

    __slots__ = ('logger_in_dist',)

    def __init__(self, system_config: Mapping):
        """Constructor."""
        super().__init__(system_config, NODE_TYPE_HUBMON)

        NET_PROTOCOL[NetProtocolID.HUBMON_LOGGING_MESSAGE].request_handler = self._handle_logging_message

        log_config = system_config['logging']

        stream_path, stream_name = log_config['log_stream'].rsplit(".", 1)
        in_dist_log_handler = logging.StreamHandler(
            stream=getattr(sys.modules[stream_path], stream_name))
        in_dist_log_handler.setLevel(log_config['log_message_level'])
        in_dist_log_handler.setFormatter(ConsoleFormatter(log_config['log_message_format_dist']))

        in_dist_logger = logging.getLogger("emews_monitor_in_dist")
        in_dist_logger.setLevel(logging.DEBUG)
        in_dist_logger.propagate = False  # don't propagate messages up to (logging) root
        in_dist_logger.addHandler(in_dist_log_handler)

        self.logger_in_dist = in_dist_logger

    def _on_scenario_start(self) -> None:
        pass

    def _on_scenario_end(self) -> None:
        pass

    def _connect_hub(self) -> None:
        """Initiate an outbound connection attempt to the hub node."""
        sys.stdout.write(f"Connecting to the hub node at {self._hub_conn.peer_address}, port {self._hub_port} ...\n")
        sys.stdout.flush()
        super()._connect_hub()

    def _on_peer_disconnection(self, peer_node_id: int, log_msg: str) -> None:
        """Handle hub node disconnection."""
        sys.stdout.write(f"\n{ConsoleFormatter.STYLE_WARNING}Disconnected from hub node: "
                         f"{log_msg}{ConsoleFormatter.STYLE_ENDC}\n")
        sys.stdout.flush()
        super()._on_peer_disconnection(peer_node_id, log_msg)

    def _on_hub_connect_failed(self, hub_address: str, reason: str) -> None:
        """Connection attempt to the hub failed."""
        sys.stdout.write(f"{ConsoleFormatter.STYLE_WARNING}Could not connect to the hub node: "
                         f"{reason}{ConsoleFormatter.STYLE_ENDC}\n")
        sys.stdout.flush()
        super()._on_hub_connect_failed(hub_address, reason)

    def _reconnect_timer(self):
        """Create a new timer to act as a delay for peer reconnection attempt."""
        sys.stdout.write(f"Waiting {self._hub_reconnect_delay} seconds before trying to connect again...\n")
        sys.stdout.flush()
        super()._reconnect_timer()

    def _on_outbound_connected(self, outbound_conn_event: OutboundEvent,
                               received_node_id: int, received_node_type: int) -> None:
        """Connection attempt to the hub node succeeded."""
        super()._on_outbound_connected(outbound_conn_event, received_node_id, received_node_type)
        sys.stdout.write(f"Connected to the hub node\n\n")
        sys.stdout.flush()

    def _handle_logging_message(self, node_id: int, src_node_id: int, src_node_name: bytes, log_msg: bytes) -> None:
        """Process the log message.  Log messages could be from any node connected to the hub."""
        assert node_id == NODE_ID_HUB

        d = pickle.loads(log_msg)
        d['args'] = None  # needed so getMessage() will not process args again
        d['exc_info'] = None  # needed so format() will not cache trace info again
        d['node_id'] = src_node_id
        d['node_name'] = src_node_name.decode()

        self.logger_in_dist.handle(logging.makeLogRecord(d))



import argparse

from emews import __version__
from emews.base.node import node_init

# TODO: test launching eMews with python -OO
# Note: log messages still use %-style, due to JIT processing of them (important to prevent overhead
# from messages which will be ignored, such as DEBUG messages when INFO is the lowest level).

parser = argparse.ArgumentParser(description='eMews network node loader')
parser.add_argument("-c", "--node_config", help="path of the node-specific configuration file")
parser.add_argument("-n", "--node_name", help="name given to this node, for example a host name")
parser.add_argument('--version', action='version', version=f'eMews {__version__}')
subparsers = parser.add_subparsers(dest='node_type', title='node_type',
                                   help='network node type to start as', required=True)

parser_host = subparsers.add_parser('host', help='host node', description='host node')
parser_host.add_argument("--hub_address", help="IPv4 address of the hub node")
parser_host.add_argument("--hub_port", type=int, help="port of the hub node")

parser_hub = subparsers.add_parser('hub', help='hub node', description='hub node')
parser_hub.add_argument("--scn_autostart", action="store_true",
                        help="start experiment scenario automatically, using values from system configuration")

parser_servspawn = subparsers.add_parser('servspawn', help='service spawner host process node',
                                         description='service spawner host process node')
parser_servspawn.add_argument("service_name", help="name of service to load on host node")
parser_servspawn.add_argument("service_configuration",
                              help="name of the service's configuration file")

parser_hubmon = subparsers.add_parser('monitor', help='hub monitor node',
                                      description='hub monitor node')
parser_hubmon.add_argument("--hub_address", help="IPv4 address of the hub node")
parser_hubmon.add_argument("--hub_port", type=int, help="port of the hub node")
parser_hubmon.add_argument("--log_local", action="store_true",
                           help="enables logging output from the monitor node itself")

parser_hubconsole = subparsers.add_parser('console', help='hub console node',
                                          description='hub console node')
parser_hubconsole.add_argument("--hub_address", help="IPv4 address of the hub node")
parser_hubconsole.add_argument("--hub_port", type=int, help="port of the hub node")
parser_hubconsole.add_argument("--log_local", action="store_true",
                               help="enables logging output from the console node itself")
args = parser.parse_args()

node_init(args).run()



"""
Throughput stats.

Outputs as log messages to the hub send/recv throughput for all interfaces on the node this service
is running on.  Currently IPv4 only.
"""
import socket

import psutil

from emews.api.service import Service


class DefaultService(Service):
    """Classdocs."""

    __slots__ = ('_if_info', '_output_interval', '_tr_unit', '_tr_scale')

    def __init__(self, config):
        """Constructor."""
        super().__init__()

        interface_prefixes = config['interface_prefixes']  # list

        output_interval = config['output_interval']
        if output_interval < 1:
            raise AttributeError("Output interval must be 1 or greater.")

        self._output_interval = output_interval

        tr_unit = config['output_unit']

        if tr_unit == 'k':
            # scale to get appropriate unit and per second
            # Kbps
            self._tr_scale = self._output_interval * 0.0078125  # 8/1024
        elif tr_unit == 'm':
            # mbps
            self._tr_scale = self._output_interval * 0.00000762939453125  # 8/(1024*1024)
        else:
            raise AttributeError(f"Output unit '{tr_unit}' invalid.")

        self._tr_unit = tr_unit
        self._if_info = {}  # stores current interface info

        ifs = psutil.net_if_addrs()
        for if_name, if_addrs in ifs.items():
            for prefix in interface_prefixes:
                if not if_name.startswith(prefix):
                    continue

                # interface is in our prefix list
                self._if_info[if_name] = {}
                # create addr string
                ipv4_str = "0.0.0.0"
                for if_addr in if_addrs:
                    # store the first IPv4 address we find, or the default if none can be found
                    if if_addr.family == socket.AF_INET:
                        ipv4_str = if_addr.address
                        self.logger.info("Monitoring NIC: %s (IPv4 address: %s)", if_name, ipv4_str)
                        break

                self._if_info[if_name]['addr_str'] = ipv4_str
                # create other map keys
                self._if_info[if_name]['b_sent'] = 0  # number of sent bytes (total)
                self._if_info[if_name]['b_recv'] = 0  # number of received bytes (total)

    def service_run(self):
        """Output throughput on all interfaces every 'output_interval' seconds."""
        # initialize the S/R byte counts to what they are currently to calibrate
        net_stats = psutil.net_io_counters(pernic=True)
        for if_name, if_info in self._if_info.items():
            if_info['b_sent'] = net_stats[if_name].bytes_sent
            if_info['b_recv'] = net_stats[if_name].bytes_recv

        self.sleep(self._output_interval)

        while not self.interrupted():
            msg_str = ""
            net_stats = psutil.net_io_counters(pernic=True)

            for if_name, if_info in self._if_info.items():
                bytes_sent = net_stats[if_name].bytes_sent
                bytes_recv = net_stats[if_name].bytes_recv

                # throughput in kbps
                tr_sent = (bytes_sent - if_info['b_sent']) * self._tr_scale
                tr_recv = (bytes_recv - if_info['b_recv']) * self._tr_scale
                msg_str = f"{msg_str}{if_info['addr_str']}: [S: {tr_sent:.2f} {self._tr_unit}bps, "\
                          f"R: {tr_recv:.2f} {self._tr_unit}bps] "
                # update the total bytes sent/recv
                if_info['b_sent'] = bytes_sent
                if_info['b_recv'] = bytes_recv

            self.logger.info(msg_str)

            # note that this is not entirely accurate, as the logic itself adds to the interval
            self.sleep(self._output_interval)



"""Example service.  Logs 'Hello world!' periodically."""

from emews.api.service import Service
from emews.environments.example.keys import Evidence


class DefaultService(Service):
    """Default service class."""

    __slots__ = ('_log_interval', '_use_env')

    def __init__(self, config):
        """Constructor."""
        super().__init__()

        self._log_interval = config['log_interval']
        self._use_env = config['use_env']

    def service_run(self):
        """Run the service."""
        self.logger.info("Hello world using an output interval of %d seconds (use environment: %s)",
                         self._log_interval, self._use_env)

        if self._use_env:
            # use example environment
            while not self.interrupted():
                # Note, the amount of time it takes to get the evidence will not be factored into
                # the sleep time, resulting in a longer interval than specified.
                global_req_count = self.ask(Evidence.HELLO_WORLD)[0]
                self.logger.info("Hello world! (global environment request count: %d)",
                                 global_req_count)
                self.sleep(self._log_interval)
        else:
            while not self.interrupted():
                self.logger.info("Hello world!")
                self.sleep(self._log_interval)


class DistWorldServer(Service):
    """Distributed hello world server."""

    __slots__ = ()

    def __init__(self, config):
        """Constructor."""
        super().__init__()

    def service_run(self):
        """Run the service."""
        self.logger.info("Listening for message ...")
        while not self.interrupted():
            # recv() blocks, unblocking on received data or if service is interrupted.  In the latter,
            # service_cleanup() will be called so the service can handle any cleanup
            sender_service_id, msg = self.recv()
            self.logger.info("Received message from service id %d: %s", sender_service_id, msg)


class DistWorldClient(Service):
    """Distributed hello world client."""

    __slots__ = ('_send_interval',)

    def __init__(self, config):
        """Constructor."""
        super().__init__()

        self._send_interval = config['send_interval']

    def service_run(self):
        """Run the service."""
        self.logger.info("Distributed hello world using a message send interval of %d seconds",
                         self._send_interval)

        while not self.interrupted():
            self.send_all("Hello world!")
            self.sleep(self._send_interval)



"""
Example environment.  Returns evidence as a request count when asked.

This environment does not implement observations.  See HelloWorld service for example of environment usage.
"""
from emews.api.env import Environment
from emews.api.env_key import EnvironmentKey

from emews.environments.example.keys import Evidence


class ExampleEnv(Environment):
    """Example environment.  Can be named anything, but only one can exist per module."""

    def __init__(self):
        """Constructor."""
        self._ev_req_count = 0  # note, no __slots__() as base Enum doesn't use them

    def get_evidence(self, node_id: int, ev_key: EnvironmentKey):
        """Return evidence based on given evidence key."""
        assert ev_key == Evidence.HELLO_WORLD

        self._ev_req_count += 1

        return self._ev_req_count,

    def update_evidence(self, node_id: int, obs_key: EnvironmentKey, obs_val) -> None:
        pass



"""Host process nodes module.  These nodes connect locally to the running host node."""
from typing import Mapping

from emews.base.node import STR_ENCODING, NODE_ID_UNASSIGNED, NODE_TYPE_HOST, NODE_TYPE_HOSTPROC, ConnectionState,\
    LocalConnectionEvent, NetProtocolID, NET_PROTOCOL, NodeBase, OutboundEvent, PeerResponseError


class HostProc(NodeBase):
    """Host node process.  Connects to local host node."""

    __slots__ = ('_port', '_reconnect_delay', '_host_conn')

    def __init__(self, system_config: Mapping):
        """Constructor."""
        super().__init__(system_config, NODE_TYPE_HOSTPROC)

        comm_config = system_config['communication']
        debug_host_port = system_config['debug']['host_port']

        if debug_host_port is not None:
            self._port = debug_host_port
        else:
            self._port = comm_config['port']

        self._reconnect_delay = comm_config['reconnect_delay']

        # HostProc doesn't process any incoming requests
        self._host_conn = ConnectionState(NODE_ID_UNASSIGNED, NODE_TYPE_HOST, comm_config['local_address'])

    def start(self) -> None:
        """Start the local process."""
        self._connect_host()

    def _on_peer_disconnection(self, peer_node_id: int, log_msg: str) -> None:
        """Handle a node disconnection."""
        self.logger.warning(log_msg)
        self._reconnect_timer()

    def _on_host_connect_failed(self, host_address: str, reason: str) -> None:
        """Connection attempt to the host failed."""
        assert host_address == self._host_conn.peer_address

        self.logger.warning("Connection attempt to the local host node at %s failed: %s", host_address, reason)
        self._reconnect_timer()

    def _on_outbound_connected(self, outbound_conn_event: OutboundEvent,
                               received_node_id: int, received_node_type: int) -> None:
        """Connection attempt to the hub succeeded."""
        host_address = self._host_conn.peer_address
        assert outbound_conn_event.peer_addr == host_address

        if received_node_type != NODE_TYPE_HOST:
            raise PeerResponseError(f"Local connection at {host_address}: expected host node type {NODE_TYPE_HOST}, "
                                    f"but received node type {received_node_type}.  Is this the local host node?")

        self._add_connection(outbound_conn_event, self._host_conn, self._on_peer_disconnection,
                             event_class=LocalConnectionEvent)

        self.logger.info("Connected to the local host node at %s (connection count: %d)",
                         host_address, self._host_conn.connection_count)

        for pending_request in self._host_conn.pending_requests:
            self._host_conn.write_data(pending_request[0])

    def _connect_host(self) -> None:
        """Initiate an outbound connection attempt to the host node."""
        self._add_outbound_connecting(self._port, self._host_conn,
                                      self._on_host_connect_failed, self._on_outbound_connected)

        self.logger.info("Connecting to the local host node at %s, port %d ...",
                         self._host_conn.peer_address, self._port)

    def _reconnect_timer(self):
        """Create a new timer to act as a delay for peer reconnection attempt."""
        timer_event = self._add_timer(self._handle_reconnect_timer)
        timer_event.set(self._reconnect_delay)
        self.logger.info("Waiting %d seconds before reconnecting ...", self._reconnect_delay)

    def _handle_reconnect_timer(self, timer_event, num_expires):
        """Handle reconnect timer expiration."""
        self._event_close(timer_event)
        # the timer should only expire once, multiple expires suggests polling loop is lagging
        self.logger.debug("Host reconnection timer expiry count: %d", num_expires)
        self._connect_host()


class ServiceSpawner(HostProc):
    """Request host to spawn service."""
    __slots__ = ('_service_name', '_service_config_filename')

    def __init__(self, system_config: Mapping, service_name: str, service_config_filename: str):
        """Constructor."""
        super().__init__(system_config)

        NET_PROTOCOL[NetProtocolID.HOST_SERVICE_LOCAL_REQ].response_handler = self._response_service_req

        self._service_name = service_name
        self._service_config_filename = service_config_filename

    def start(self) -> None:
        """Start the local process."""
        self.logger.info("Service to spawn: %s, configuration: %s",
                         self._service_name, self._service_config_filename)
        super().start()
        self._send_net_request(self._host_conn, NetProtocolID.HOST_SERVICE_LOCAL_REQ,
                               (bytes(self._service_name, STR_ENCODING),
                                bytes(self._service_config_filename, STR_ENCODING)))

    def _response_service_req(self, peer_node_id: int, local_data):
        """Ack from host to acknowledge service request."""
        self.logger.info("Service spawn request acknowledged, shutting down process...")
        self.stop()



"""Host client node module."""
from typing import Dict, Mapping, Set, Tuple, Type

import struct
import threading

from emews.base.client import SynchronizedClientNode
from emews.base.env_key import EnvironmentKey
from emews.base.node import NET_PROTOCOL, NetProtocolID, NODE_ID_HUB, NODE_ID_UNASSIGNED, NODE_TYPE_HOST,\
    NODE_TYPE_HOSTPROC, TYPE_LEN, STR_ENCODING, NetworkHandlerError, InboundEvent, ConnectionState, LocalConnectionEvent
from emews.base.service import load_service, ServiceControlError, ServiceLoadError, Service, ServiceImpl

TYPE_LEN_INT = TYPE_LEN['L']


class HostNode(SynchronizedClientNode):
    """Host client node.  These nodes run services."""

    __slots__ = ('_port', '_services', '_timeout_service_shutdown', '_env_keys', '_env_key_pending', '_env_key_cv',
                 '_is_scn_started')

    def __init__(self, system_config: Mapping):
        """Constructor."""
        super().__init__(system_config, NODE_TYPE_HOST)

        NET_PROTOCOL[NetProtocolID.HOST_SERVICE_SEND_MSG].request_handler = self._handle_service_recv_msg
        NET_PROTOCOL[NetProtocolID.HOST_SERVICE_SPAWN].request_handler = self._handle_cmd_service_spawn
        NET_PROTOCOL[NetProtocolID.HOST_SERVICE_START].request_handler = self._handle_cmd_service_start
        NET_PROTOCOL[NetProtocolID.HOST_SERVICE_STOP].request_handler = self._handle_cmd_service_stop
        NET_PROTOCOL[NetProtocolID.HOST_SERVICE_LOCAL_REQ].request_handler = self._handle_local_service_spawn

        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_REG].response_handler = self._response_hub_service_reg
        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_SEND_ALL].response_handler = self._response_unblock_service
        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_SEND_TO].response_handler = self._response_unblock_service
        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_TELL].response_handler = self._response_unblock_service
        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_ASK].response_handler = self._response_env_ask
        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_ENVKEY_ID].response_handler = self._response_env_key_id
        NET_PROTOCOL[NetProtocolID.HUB_SERVICE_SPAWNED].response_handler = self._response_hub_service_spawned

        debug_config = system_config['debug']
        debug_host_port = debug_config['host_port']

        if debug_host_port is not None:
            self._port = debug_host_port
        else:
            self._port = self._hub_port

        self._services: Dict[int, ServiceImpl] = {}  # all services (key=service id)

        self._env_keys: Dict[EnvironmentKey, int] = {}  # cached key enum -> key id
        self._env_key_pending: Set[EnvironmentKey] = set()  # key id requests not received yet
        self._env_key_cv = threading.Condition()  # sync access to env_key dict

        self._is_scn_started = False

        timeout_service_shutdown = system_config['general']['service_shutdown_timeout']
        if timeout_service_shutdown is None or timeout_service_shutdown < 1:
            self._timeout_service_shutdown = 1
        else:
            self._timeout_service_shutdown = timeout_service_shutdown

    def start(self) -> None:
        """Start the node."""
        if self._add_listener(self._port, self._on_inbound_connected):
            super().start()

    def shutdown(self) -> None:
        """Shutdown all services."""
        self.logger.info("Will wait up to %d seconds for each service to stop", self._timeout_service_shutdown)

        active_services = []
        for service in self._services.values():
            if service.is_active():
                try:
                    service.exit()
                except ServiceControlError as ex:
                    err_msg = f"On attempted service exit: {ex}"
                    self.logger.warning(err_msg)
                    self.logger_dist.warning(err_msg)
                else:
                    service.service_thread.join(self._timeout_service_shutdown)
                    if service.is_active():
                        # service is still running, join timeout reached
                        active_services.append(service.name)

        if len(active_services):
            self.logger.warning("%d services are still running: [%s], continuing shutdown",
                                len(active_services), ", ".join(active_services))
        else:
            self.logger.info("All running services shutdown")

    def _on_local_disconnection(self, peer_node_id: int, log_msg: str) -> None:
        """Handle a local disconnection."""
        assert peer_node_id == NODE_ID_UNASSIGNED
        self.logger.debug(log_msg)

    def _on_inbound_connected(self, inbound_conn_event: InboundEvent,
                              received_node_name: str, received_node_type: int, received_num_conn: int) -> None:
        """Handle incoming local connection for local requests."""
        peer_addr = inbound_conn_event.peer_addr
        if not peer_addr.startswith('127.0.'):
            self.logger.warning("Incoming connection is non-local (address: %s, received node name: %s, "
                                "received node type: %d), dropping ...",
                                peer_addr, received_node_name, received_node_type)
            self._event_close(inbound_conn_event)
            return

        if received_node_type != NODE_TYPE_HOSTPROC:
            self.logger.warning("Incoming connection at %s: Expected node type %d, instead received node type %d",
                                peer_addr, NODE_TYPE_HOSTPROC, received_node_type)
            self._event_close(inbound_conn_event)
            return

        local_conn_state = ConnectionState(NODE_ID_UNASSIGNED, received_node_type, peer_addr)
        self._add_connection(inbound_conn_event, local_conn_state, self._on_local_disconnection,
                             event_class=LocalConnectionEvent)

        self.logger.info("Local connection at %s: host process '%s' connected", peer_addr, received_node_name)

        # send our node id and type
        # TODO: If HostProc needs our node id, wait to send until we have our node id from hub
        local_conn_state.write_data(struct.pack('>LH', self.node_id, self.node_type))

    def _service_start(self, service: ServiceImpl) -> None:
        """Start the service given."""
        service_name = service.name

        info_msg = f"Starting service '{service_name}' ..."
        self.logger.info(info_msg)
        self.logger_dist.info(info_msg)

        try:
            service.start()
        except ServiceControlError as ex:
            err_msg = f"On start for service '{service_name}': {ex}"
            self.logger.warning(err_msg)
            self.logger_dist.warning(err_msg)

    def _service_stop(self, service: ServiceImpl) -> None:
        """Stop the service given."""
        service_name = service.name

        info_msg = f"Stopping service '{service_name}' ..."
        self.logger.info(info_msg)
        self.logger_dist.info(info_msg)

        try:
            service.stop()
        except ServiceControlError as ex:
            err_msg = f"On stop for service '{service_name}': {ex}"
            self.logger.warning(err_msg)
            self.logger_dist.warning(err_msg)

    def _on_scenario_start(self) -> None:
        """Start all services."""
        self._is_scn_started = True
        self.logger.info("Experiment scenario has started, starting %d services ...", len(self._services))
        self.logger_dist.info("Starting %d services ...", len(self._services))

        for service in self._services.values():
            self._service_start(service)

    def _on_scenario_end(self) -> None:
        """Stop all services."""
        self._is_scn_started = False

        running_services = []

        for service in self._services.values():
            if service.is_running():
                running_services.append(service)

        self.logger.info("Experiment scenario has ended, stopping %d running services ...", len(self._services))
        self.logger_dist.info("Stopping %d running services ...", len(running_services))

        for service in running_services:
            self._service_stop(service)

    ##################################################
    # synchronized methods (these run from any thread)
    ##################################################
    def _sync_get_env_key_id(self, env_key: EnvironmentKey) -> int:
        """
        Fetch the key id given key enum.

        This method will always return the key id, blocking if the id needs to be fetched.
        """
        with self._env_key_cv:
            try:
                env_key_id = self._env_keys[env_key]
            except KeyError:
                if env_key not in self._env_key_pending:
                    # we need to request the key id from the hub
                    self._env_key_pending.add(env_key)
                    self._send_hub_request(NetProtocolID.HUB_SERVICE_ENVKEY_ID,
                                           request_data=(bytes(env_key.key_full_name(), STR_ENCODING),),
                                           local_data=env_key)

                    self.logger.debug("Requesting key id for environment key '%s' (%s)",
                                      env_key.key_full_name(), str(env_key))

                # block the service thread until the key id is fetched, then return it
                while env_key in self._env_key_pending:
                    # Releases lock on wait, acquires it when done waiting.  Note that if multiple
                    # keys are being requested, this service may awaken from another fulfilled
                    # request, in which case 'env_key' will still be in the pending set, and this
                    # service will continue to wait.
                    self._env_key_cv.wait()

                env_key_id = self._env_keys[env_key]

            return env_key_id

    def _sync_service_init(self, service: ServiceImpl, ex: Exception = None, exc_trace: str = None) -> None:
        """Service instantiation result."""
        self._sync_add_task(self._task_service_init, service, ex, exc_trace)

    def _sync_service_exit(self, service: ServiceImpl, ex: Exception = None, exc_trace: str = None) -> None:
        """Service exit result."""
        self._sync_add_task(self._task_service_exit, service, ex, exc_trace)

    #############################################
    # thread tasks (these run in the main thread)
    #############################################
    def _task_service_init(self, service: ServiceImpl, ex: Exception = None, exc_trace: str = None) -> None:
        """Service has finished instantiating (or failed to instantiate)."""
        if ex is not None:
            self.logger.error("Service '%s' did not instantiate: %s\n%s", service.name, ex, exc_trace)
        else:
            self.logger.info("Service '%s' instantiated", service.name)

        self._send_hub_request(NetProtocolID.HUB_SERVICE_SPAWNED, request_data=(service.service_id, ex is None),
                               local_data=service)

    def _task_service_exit(self, service: ServiceImpl, ex: Exception = None, exc_trace: str = None) -> None:
        """Service thread has terminated."""
        if ex is not None:
            self.logger.error("Service '%s' terminated unexpectedly: %s\n%s", service.name, ex, exc_trace)
        else:
            self.logger.info("Service '%s' shutdown successful", service.name)

        self._send_hub_request(NetProtocolID.HUB_SERVICE_EXITED, request_data=(service.service_id, ex is not None))

    ##########################
    # network request handlers
    ##########################
    def _handle_service_recv_msg(self, node_id: int, from_service_id: int, to_service_id: int, data: bytes) -> tuple:
        """Delegate received data to service instance with service id."""
        assert node_id == NODE_ID_HUB

        try:
            service = self._services[to_service_id]
        except KeyError:
            # indicates a bug in the hub
            self.logger_dist.error("Service with id %d is not registered to us", to_service_id)
            return ()

        service.recv_data(from_service_id, data)

        return ()

    def _handle_cmd_service_start(self, node_id: int, service_id: int) -> tuple:
        """Start a service registered to us."""
        assert node_id == NODE_ID_HUB

        try:
            service = self._services[service_id]
        except KeyError:
            # indicates a bug in the hub
            self.logger_dist.error("Service with id %d is not registered to us", service_id)
        else:
            self._service_start(service)

        return ()

    def _handle_cmd_service_stop(self, node_id: int, service_id: int) -> tuple:
        """Stop a service registered to us."""
        assert node_id == NODE_ID_HUB

        try:
            service = self._services[service_id]
        except KeyError:
            # indicates a bug in the hub
            self.logger_dist.warning("Service with id %d is not registered to us", service_id)
        else:
            self._service_stop(service)

        return ()

    def _handle_cmd_service_spawn(self, node_id: int, service_name: bytes, config_filename: bytes,
                                  service_id: int, service_display_name: bytes) -> tuple:
        """Spawn a service, requested by the hub (probably from a console)."""
        assert node_id == NODE_ID_HUB

        service_name_str = service_name.decode()
        config_filename_str = config_filename.decode()

        try:
            service_class, service_config = load_service(service_name_str, config_filename_str)
        except ServiceLoadError as ex:
            raise NetworkHandlerError(ex) from ex

        self.logger.info("New service spawn remote request for service '%s' with configuration '%s'",
                         service_name_str, config_filename_str)

        # instantiate using the response method for service reg
        self._response_hub_service_reg(node_id,
                                       (service_class, service_name_str, service_config, config_filename_str),
                                       service_id, service_display_name)

        return ()

    def _handle_local_service_spawn(self, node_id: int, service_name: bytes, config_filename: bytes) -> tuple:
        """Import a service.  Request service information from hub."""
        assert node_id == NODE_ID_UNASSIGNED  # from local service request process

        service_name_str = service_name.decode()
        config_filename_str = config_filename.decode()

        try:
            service_class, service_config = load_service(service_name_str, config_filename_str)
        except ServiceLoadError as ex:
            raise NetworkHandlerError(ex) from ex

        self.logger.info("New service spawn request for service '%s' with configuration '%s'",
                         service_name_str, config_filename_str)

        self._send_hub_request(NetProtocolID.HUB_SERVICE_REG,
                               request_data=(service_name, config_filename),
                               local_data=(service_class, service_name_str, service_config, config_filename_str))

        return ()

    ###########################
    # network response handlers
    ###########################
    def _response_hub_service_reg(self, node_id: int, local_data: Tuple[Type[Service], str, Dict, str],
                                  service_id: int, service_display_name: bytes) -> None:
        """Complete service initialization (spawn)."""
        assert node_id == NODE_ID_HUB
        service_class, service_name, service_config, config_filename = local_data

        service_logger = self._get_dist_logger(f"emews_service.{service_id}")
        service_full_name = f"[{service_id}]{service_display_name.decode()}.{config_filename}"

        info_msg = f"Instantiating service '{service_name}' (assigned id: {service_id}) as '{service_full_name}' ..."
        self.logger.info(info_msg)
        self.logger_dist.info(info_msg)

        service = ServiceImpl(self.logger_dist, service_id, service_full_name,
                              (service_logger, service_class, service_config,
                               self._sync_service_init, self._sync_service_exit),
                              self._sync_get_env_key_id,
                              self._send_hub_request)

        self._services[service_id] = service

        service.service_thread.start()

    def _response_hub_service_spawned(self, node_id: int, service: ServiceImpl) -> None:
        """ACK from hub node of service spawned / instantiation result."""
        assert node_id == NODE_ID_HUB

        if self._is_scn_started:
            self._service_start(service)

    @staticmethod
    def _response_unblock_service(node_id: int, service_event: threading.Event):
        """Hub response from a blocking service request."""
        assert node_id == NODE_ID_HUB

        service_event.set()  # unblock service that sent request

    @staticmethod
    def _response_env_ask(node_id: int, local_data: Tuple[threading.Event, EnvironmentKey, list], env_data: bytes):
        """Env data from a service ask request."""
        assert node_id == NODE_ID_HUB

        service_event, ev_key, ev_lst = local_data

        assert not len(ev_lst)

        num_ev_val: int = struct.unpack_from(">L", env_data)[0]
        ev_lst.extend(struct.unpack_from(f">{ev_key.key_list_type * num_ev_val}", env_data, offset=TYPE_LEN_INT))
        service_event.set()  # unblock service (service has reference to ev_lst)

    def _response_env_key_id(self, node_id: int, env_key: EnvironmentKey, env_key_id: int):
        """Environment key id from a env key request."""
        assert node_id == NODE_ID_HUB

        with self._env_key_cv:
            self._env_keys[env_key] = env_key_id
            self._env_key_pending.remove(env_key)

            # This will wake up any services waiting for any key request.  Those not waiting on this
            # specific key will continue to wait after checking the key pending set.
            self._env_key_cv.notify_all()

        self.logger.debug("Environment key '%s' has assigned key id %d", env_key.key_full_name(), env_key_id)



"""Web crawler with environment support for link selection based on popularity."""
from typing import AnyStr, List, Set, Tuple, Union

import logging
import math
import random
import socket
import struct

import bs4
from urllib.parse import urljoin, urlparse
import urllib3

from emews.environments.web.keys import Observation, Evidence
from emews.api.service import Service
from emews.api.random import TruncnormInt, UniformInt


class AgentModel:
    """Next link sampler."""

    __slots__ = ('logger', 'prior', 'ev_viral', 'ev_visited', 'num_pref_links')

    def __init__(self, logger: logging.Logger):
        """Constructor."""
        self.logger = logger

        self.prior: List[float] = [0.0] * 2  # class variable
        self.ev_viral: List[float] = [0.0] * 2  # attribute variables
        self.ev_visited: List[float] = [0.0] * 2
        self.num_pref_links = 0

    def _set_prior_param(self, num_links: int, links_preferred: List[int], links_preferred_strength: float):
        """Set the prior."""
        num_pref_links = 0
        for index in links_preferred:
            if 0 <= index < num_links:
                num_pref_links += 1

        norm = ((num_pref_links * links_preferred_strength) + num_links - num_pref_links) / float(num_links)

        self.num_pref_links = num_pref_links
        self.prior[0] = links_preferred_strength / (num_links * norm)
        self.prior[1] = 1.0 / (num_links * norm)

    def _set_evidence_params(self, num_links: int, link_viral_strength: float, link_visited_strength: float):
        """Set the evidence variable parameters."""
        norm = (link_viral_strength + num_links - 1.0) / float(num_links)
        self.ev_viral[0] = link_viral_strength / (num_links * norm)
        self.ev_viral[1] = 1.0 / (num_links * norm)

        norm = (link_visited_strength + num_links - 1.0) / float(num_links)
        self.ev_visited[0] = link_visited_strength / (num_links * norm)
        self.ev_visited[1] = 1.0 / (num_links * norm)

    @staticmethod
    def _prob_to_log(prob: float) -> float:
        """Convert a probability to a log prob."""
        if prob > 0.0:
            return math.log(prob)  # base e

        return -28.0  # represent an arbitarily small probability

    def sample(self, num_links: int, links_preferred: List[int], visited_links: Set[int],
               viral_links: Tuple[int]) -> int:
        """Perform inference on the factored joint, and sample."""
        prob_to_log = self._prob_to_log
        log_prior_pref = prob_to_log(self.prior[0])
        log_prior_nonpref = prob_to_log(self.prior[1])
        log_viral = prob_to_log(self.ev_viral[0])
        log_nonviral = prob_to_log(self.ev_viral[1])
        log_visited = prob_to_log(self.ev_visited[0])
        log_nonvisited = prob_to_log(self.ev_visited[1])

        posteriors: List[float] = []
        # init to appropriate prior
        for i in range(num_links):
            if i in links_preferred:
                posteriors.append(log_prior_pref)
            else:
                posteriors.append(log_prior_nonpref)

        for i in range(num_links):
            # for each link in prior (links to click)
            nonviral_count = num_links
            nonvisited_count = num_links
            if i in viral_links:
                # if the link at index i is viral, append viral prob for evidence node i
                # Note that we only do this once per index i, even if multiple videos are viral.
                # This is a consequence of the naive Bayes assumption.
                posteriors[i] += log_viral
                nonviral_count -= 1  # the link at index i is viral, so decrement
            if i in visited_links:
                posteriors[i] += log_visited
                nonvisited_count -= 1

            posteriors[i] += nonviral_count * log_nonviral
            posteriors[i] += nonvisited_count * log_nonvisited

        # find the normalization constant, P(e).  Because we are in log space we gotta be sneaky.
        # First, find the smallest log and pre-normalize the others with it, then, exponentiate
        smallest_posterior = min(posteriors)

        for i in range(len(posteriors)):
            posteriors[i] -= smallest_posterior
            if posteriors[i] < -21:
                # anything smaller than this may cause underflow issues, treat as exp(zero) = 1
                posteriors[i] = 1.0
            else:
                posteriors[i] = math.exp(posteriors[i])

        norm = 1.0 / math.fsum(posteriors)  # p(e)

        # normalize posteriors
        for i in range(len(posteriors)):
            posteriors[i] = posteriors[i] * norm

        self.logger.debug("Agent Model: posterior: %s", posteriors)

        # sample link index from the posterior CDF
        r_sample = random.random()  # [0.0, 1.0)
        cdf_posterior = 0.0
        selected_link_index = 0
        for posterior in posteriors:
            cdf_posterior += posterior
            if cdf_posterior > r_sample:
                # sample is within this interval, corresponds to the link index
                break
            selected_link_index += 1

        # self.logger.debug("SAMPLE: link index (%d/%d), random_sample: %f", selected_link_index, num_links, r_sample)
        return selected_link_index

    def update_model(self, num_links: int, links_preferred: List[int], links_preferred_strength: float,
                     link_viral_strength: float, link_visited_strength: float):
        """Build the model based on the number of links.  Evidence count is consistent."""
        # rebuild prior (class) distribution
        self._set_prior_param(num_links, links_preferred, links_preferred_strength)
        # rebuild conditional distributions
        self._set_evidence_params(num_links, link_viral_strength, link_visited_strength)
        self.logger.debug("Agent Model: attributes (prior): pref=%f, not_pref=%f "
                          "(state_space=2, num_attributes=%d, num_pref_links=%d)",
                          self.prior[0], self.prior[1], num_links, self.num_pref_links)
        self.logger.debug(
            "Agent Model: attributes (evidence): viral=%f, not_viral=%f (state_space=2, num_attributes=%d)",
            self.ev_viral[0], self.ev_viral[1], num_links)
        self.logger.debug(
            "Agent Model: attributes (evidence): visited=%f, not_visited=%f (state_space=2, num_attributes=%d)",
            self.ev_visited[0], self.ev_visited[1], num_links)


class BaseCrawler(Service):
    """Base class for web crawling clients."""

    __slots__ = ('_http',
                 '_invalid_link_prefixes',
                 '_siteURLs',
                 '_std_deviation_link',
                 '_site_sampler',
                 '_crawl_sampler',
                 '_num_links_sampler',
                 '_link_delay_sampler')

    def __init__(self, config: dict):
        """Constructor."""
        super().__init__()

        self._invalid_link_prefixes: List[str] = config['invalid_link_prefixes']
        self._siteURLs: List[str] = config['start_sites']

        self._site_sampler = UniformInt(upper_bound=len(self._siteURLs) - 1)
        self._num_links_sampler = UniformInt(**config['num_links_sampler'])
        self._crawl_sampler = UniformInt(**config['crawl_sampler'])
        self._link_delay_sampler = UniformInt(**config['link_delay_sampler'])

        # urllib3 related setup
        urllib3.disable_warnings()  # ignore self-signed SSL certs
        br_header = urllib3.util.make_headers(user_agent=config['user_agent'])
        self._http = urllib3.PoolManager(num_pools=2, headers=br_header, cert_reqs='CERT_NONE')

    def get_next_link_index(self, page_links: List[str]) -> int:
        """Given a list of page links, find and return the index of the first valid link."""
        raise NotImplementedError

    def service_run(self) -> None:
        """Run the service."""
        while not self.interrupted():
            self.sleep(self._crawl_sampler.sample())
            self._crawl(self._siteURLs[self._site_sampler.sample()])

    def _checklink(self, link_str: Union[str, None]) -> bool:
        """Check a link object's URL for validity (not a javascript link or something)."""
        if link_str is None:
            return False

        for prefix in self._invalid_link_prefixes:
            if link_str.lower().startswith(prefix):
                return False

        return True

    def _get_page_links(self, site_url: str, raw_html: AnyStr) -> List[str]:
        """Parse a list of links from the raw HTML."""
        soup_html = bs4.BeautifulSoup(raw_html, 'html.parser')  # using standard python html parser

        page_links: List[str] = []
        for page_link in soup_html.find_all('a'):
            if self._checklink(page_link.get('href')):
                page_links.append(urljoin(site_url, page_link.get('href')))

        return page_links

    def _crawl(self, site_host):
        """Web crawl, starting from the _siteURL, picking links from each page visited."""
        num_links_to_crawl = self._num_links_sampler.sample()

        self.logger.info("Will crawl maximum of %d pages", num_links_to_crawl)

        page_url = site_host
        crawl_str = f"Starting crawl at {site_host}"

        num_pages_visited = 0

        # now crawl for (max) num_links_to_crawl
        for _ in range(num_links_to_crawl):
            self.sleep(self._link_delay_sampler.sample())
            self.logger.info(crawl_str)

            try:
                site_page = self._http.request('GET', page_url).data
            except Exception as ex:
                self.logger.warning("On site open: %s, (server: %s)", ex, site_host)
                return

            num_pages_visited += 1
            site_page_links = self._get_page_links(site_host, site_page)

            if not site_page_links:
                self.logger.debug("Crawled page '%s' doesn't have any links to click.", page_url)
                break

            selected_link_index = self.get_next_link_index(site_page_links)

            page_url = site_page_links[selected_link_index]
            link_split = urlparse(bytes(page_url, 'utf_8'))
            site_host: str = f"{link_split.scheme.decode()}://{link_split.netloc.decode()}"

            crawl_str = f"Clicked link [{selected_link_index}/{len(site_page_links) - 1}]: {page_url}"

        self.logger.info("Finished crawl at %s (pages visited: %d)", site_host, num_pages_visited)


class DefaultService(BaseCrawler):
    """Web crawler client."""

    __slots__ = ('_first_page',
                 '_next_link_sigma_first',
                 '_next_link_sigma')

    def __init__(self, config: dict):
        """Constructor."""
        super().__init__(config)

        self._first_page = True
        self._next_link_sigma_first: float = config['next_link_sigma_first']
        self._next_link_sigma: float = config['next_link_sigma']

    def get_next_link_index(self, page_links: List[str]) -> int:
        """Given a list of page links, find and return the index of the first valid link."""
        if self._first_page:
            self._first_page = False
            std_deviation = self._next_link_sigma_first
        else:
            std_deviation = self._next_link_sigma

        if len(page_links) == 1:
            return 0

        selected_link_index = TruncnormInt(upper_bound=len(page_links) - 1, sigma=std_deviation).sample()

        return selected_link_index


class AgentCrawler(BaseCrawler):
    """Web crawler that interacts with its environment."""

    __slots__ = ('_link_viral_strength',
                 '_link_visited_strength',
                 '_next_link_model',
                 '_links_preferred',
                 '_links_preferred_strength',
                 '_visited_links')

    def __init__(self, config: dict):
        """Constructor."""
        super().__init__(config)

        self._links_preferred: List[int] = config['links_preferred']

        self._links_preferred_strength: float = config['links_preferred_strength']
        self._link_viral_strength: float = config['link_viral_strength']
        self._link_visited_strength: float = config['link_visited_strength']

        self._next_link_model = AgentModel(self.logger)

        self._visited_links: Set[int] = set()  # local evidence

    def service_run(self) -> None:
        """Run the service."""
        while not self.interrupted():
            self.sleep(self._crawl_sampler.sample())

            site_host = self._siteURLs[self._site_sampler.sample()]
            # Important: the agent assumes the site visited doesn't change, ie, video hosting site.
            # The site to crawl should not contain external links.
            crawl_ip: Tuple = struct.unpack(">I", socket.inet_aton(site_host.rsplit('/', 1)[-1]))
            self.tell(Observation.CRAWL_SITE, crawl_ip)  # tell environment the site we are going to crawl

            self._crawl(site_host)

            self._visited_links.clear()

    def get_next_link_index(self, page_links: List[str]) -> int:
        """Given a list of page links, find and return the index of the first valid link."""
        self._next_link_model.update_model(len(page_links), self._links_preferred, self._links_preferred_strength,
                                           self._link_viral_strength, self._link_visited_strength)

        ev_visited_links = self._visited_links
        ev_viral_links = self.ask(Evidence.VIRAL_LINKS)

        self.logger.debug("Current evidence (viral link indices): %s", ev_viral_links)

        selected_link_index = self._next_link_model.sample(len(page_links), self._links_preferred, ev_visited_links,
                                                           ev_viral_links)

        self.tell(Observation.LINK_CLICKED, (selected_link_index,))
        ev_visited_links.add(selected_link_index)

        return selected_link_index



"""
Environment keys for example environment.

This is a separate module to enable key importing without having to indirectly import env.  Only the hub needs to
import the environment module (env), not services running on hosts.
"""

from emews.api.env_key import EnvironmentKey


class Evidence(EnvironmentKey):
    """Evidence environment key enums.  Evidence key class must be named 'Evidence'."""
    HELLO_WORLD = int  # list type (observations and evidence can take zero or more values in a list)



"""Environment for SiteCrawler.  Provides evidence for viral links."""
from typing import Dict, List, Set, Tuple

import socket
import struct

from emews.api.env import Environment
from emews.api.env_key import EnvironmentKey

from emews.environments.web.keys import Observation, Evidence


class WebEnv(Environment):
    """Classdocs."""

    __slots__ = ('_site_map', '_link_data', '_timer_data', '_viral_links', '_viral_link_threshold',
                 '_viral_link_expiration')

    # enums
    LINK_COUNT = 0
    LINK_VIRAL = 1

    def __init__(self):
        """Constructor."""
        super().__init__()

        self._site_map: Dict[int, int] = {}  # [node_id]: current site
        self._link_data: Dict[int, Dict[int, List]] = {}  # [site]: {[link_index]: link_data}
        self._timer_data: Dict[int, Tuple] = {}  # [timer_fd]: timer data
        self._viral_links: Dict[int, Set[int]] = {}  # [site]: set of viral links

        # viral link parameters
        self._viral_link_threshold = 10  # number of specific links clicks, per site, before viral
        self._viral_link_expiration = 60  # seconds that a link remains viral

    def update_evidence(self, node_id: int, obs_key: EnvironmentKey, obs_val) -> None:
        """Update evidence given new observation."""
        assert len(obs_val) == 1
        if obs_key == Observation.CRAWL_SITE:
            # node is starting a new crawl
            self._site_map[node_id] = obs_val[0]
        elif obs_key == Observation.LINK_CLICKED:
            # node clicked on a link
            self._update_clicked_links(node_id, obs_val[0])

    def get_evidence(self, node_id: int, ev_key: EnvironmentKey):
        """Return the relevant list of evidence given the key and a node id."""
        assert ev_key == Evidence.VIRAL_LINKS

        crawl_site = self._site_map.get(node_id)
        if crawl_site is None:
            return ()

        viral_links = self._viral_links.get(crawl_site)
        if viral_links is None:
            return ()

        return viral_links

    def _update_clicked_links(self, node_id: int, clicked_link: int):
        """Update relevant evidence in regard to a new click."""
        # New_obs.val is the link index clicked on.  Simply check if enough clicks have occurred.
        # The agent needs to send an observation on what site it is crawling before sending link clicks
        crawl_site = self._site_map[node_id]

        if crawl_site not in self._link_data:
            self._link_data[crawl_site] = {}

        link_data = self._link_data[crawl_site]  # link data is for a specific site

        if clicked_link not in link_data:
            link_data[clicked_link] = [0, False]  # [link index]: number of clicks, went viral?

        per_link_data = link_data[clicked_link]

        num_clicks = per_link_data[WebEnv.LINK_COUNT] + 1
        per_link_data[WebEnv.LINK_COUNT] = num_clicks

        if num_clicks > self._viral_link_threshold and not per_link_data[WebEnv.LINK_VIRAL]:
            # viral link, update evidence (used for agent ask)
            per_link_data[WebEnv.LINK_VIRAL] = True
            if crawl_site not in self._viral_links:
                self._viral_links[crawl_site] = set()

            self._viral_links[crawl_site].add(clicked_link)

            self.logger.info("link on server '%s' at index %d has gone viral",
                             socket.inet_ntoa(struct.pack(">I", crawl_site)), clicked_link)

            timer_fd = self.new_timer(self._viral_link_expiration, self._evidence_viral_link_expired, repeat=False)
            self._timer_data[timer_fd] = (crawl_site, clicked_link)

    def _evidence_viral_link_expired(self, timer_fd: int, num_expires: int):
        """When a timer has finished, this will be invoked."""
        crawl_site, link_index = self._timer_data[timer_fd]
        self._viral_links[crawl_site].remove(link_index)

        self.del_timer(timer_fd)

        self.logger.info("link on server '%s' at index '%d' is no longer viral",
                         socket.inet_ntoa(struct.pack(">I", crawl_site)), link_index)



"""Base module for eMews services."""
from typing import Any, Callable, Deque, Dict, List, Mapping, Set, Tuple, Type, Union

import collections
import importlib
import logging
import pickle
import struct
import traceback

from threading import Condition, Event, Lock, Thread
from time import perf_counter

from emews.base.config import parse_service_config, get_service_class_name
from emews.base.env_key import EnvironmentKey, ENV_KEY_TYPE
from emews.base.node import NetProtocolID


class ServiceInterrupted(Exception):
    """Raised when a service is interrupted and a service call is invoked or service blocked."""
    pass


class ServiceLoadError(Exception):
    """Raised when a service cannot be imported."""
    pass


class ServiceControlError(Exception):
    """Raised when a service command (start/stop/exit) fails."""
    pass


# noinspection PyUnusedLocal
def no_call_interrupted(*args, **kwargs):
    raise ServiceInterrupted


def load_service(service_name: str, config_filename: str) -> Tuple[Type['Service'], Dict]:
    """Import a service class and configuration."""
    try:
        service_config = parse_service_config(service_name, config_filename)
    except Exception as ex:
        # the parser is dependent on third party YAML parsers, catch all exceptions here and re-raise
        raise ServiceLoadError(f"Service configuration for service '{service_name}' could not be loaded: {ex}") from ex

    if 'parameters' not in service_config:
        raise ServiceLoadError(
            f"Service configuration for service '{service_name}' missing required section: 'parameters'")

    service_class_str = get_service_class_name(service_config)

    try:
        service_class = getattr(importlib.import_module(f"emews.services.{service_name}.service"),
                                service_class_str)
    except (AttributeError, ImportError, SyntaxError) as ex:
        raise ServiceLoadError(
            f"Service '{service_name}' (service class: {service_class_str}) could not be loaded: {ex}") from ex

    return service_class, service_config


class Service:
    """Service base class."""

    __slots__ = ('_logger', '_cb_service_call',)

    def __init_subclass__(cls, **kwargs):
        """Subclass creation hook.  'cls' is the subclass (child Service)."""
        super().__init_subclass__(**kwargs)
        service_init = cls.__init__  # subclass init will have a different signature (config arg)

        def __init__(self: Service, config: Mapping, *init_args, **init_kwargs):
            """Child service __init__ override."""
            if not hasattr(self, '_logger'):
                # this will be true only for the concrete child class (last one in inheritance chain)
                assert len(init_args) == 2 and not len(init_kwargs)
                self._logger = init_args[0]  # unique logger to service instance
                self._cb_service_call = init_args[1]

                service_init(self, config)
                return

            # config is required, anything else is optional, passed by child in super().__init__()
            service_init(self, config, *init_args, **init_kwargs)

        setattr(cls, '__init__', __init__)  # replace child class init with the def above

    def service_run(self) -> None:
        """Point of entry for running the service."""
        raise NotImplementedError

    def service_exit(self) -> None:
        """Called when a service exits, either by interruption or service_run() exit.  Handle any cleanup here."""
        pass

    @property
    def logger(self) -> logging.Logger:
        """System logger."""
        return self._logger

    def interrupted(self) -> bool:
        """Return interrupted flag.  If True, service should quickly stop its execution."""
        return self._cb_service_call(ServiceImpl.CALL_INDEX_INTERRUPTED)

    def elapsed_time(self) -> float:
        """Return the elapsed time from start of service run."""
        return self._cb_service_call(ServiceImpl.CALL_INDEX_ELAPSED_TIME)

    def sleep(self, duration: Union[float, int]) -> None:
        """Suspend execution of service for the given amount of time (in seconds)."""
        if duration < 0 or duration is None:
            duration = 0
        self._cb_service_call(ServiceImpl.CALL_INDEX_SLEEP, duration)

    def send_all(self, data: Any) -> None:
        """Send data to all running services that are designated as receivers."""
        self._cb_service_call(ServiceImpl.CALL_INDEX_SEND_ALL, data)

    def send_to(self, data: Any, *to_service_ids: int) -> None:
        """Send data to all running services in to_service_ids that are designated as receivers."""
        self._cb_service_call(ServiceImpl.CALL_INDEX_SEND_TO, to_service_ids, data)

    def recv(self) -> Tuple[int, Any]:
        """Wait until data is received through the Hub.  Returns the service id which sent the data, and the data."""
        return self._cb_service_call(ServiceImpl.CALL_INDEX_RECV_MSG)

    def received_data(self) -> bool:
        """Returns True if there is data waiting to be fetched using recv()."""
        return self._cb_service_call(ServiceImpl.CALL_INDEX_HAS_RECV)

    def tell(self, obs_key: EnvironmentKey, obs_val: ENV_KEY_TYPE) -> None:
        """Give an observation to the network environment."""
        self._cb_service_call(ServiceImpl.CALL_INDEX_TELL, obs_key, obs_val)

    def ask(self, ev_key: EnvironmentKey) -> ENV_KEY_TYPE:
        """Ask evidence from the network environment."""
        return self._cb_service_call(ServiceImpl.CALL_INDEX_ASK, ev_key)


class ServiceImpl:
    """Service implementation."""

    CALL_INDEX_INTERRUPTED = 0
    CALL_INDEX_SLEEP = 1
    CALL_INDEX_SEND_ALL = 2
    CALL_INDEX_SEND_TO = 3
    CALL_INDEX_RECV_MSG = 4
    CALL_INDEX_HAS_RECV = 5
    CALL_INDEX_TELL = 6
    CALL_INDEX_ASK = 7
    CALL_INDEX_ELAPSED_TIME = 8

    __slots__ = ('logger_dist', 'service_id', 'service_thread', '_thread_event', '_start_time', '_is_running',
                 '_do_exit', '_service_calls', '_interrupted', '_events', '_event_set_lock', '_cv',
                 '_recv_queue', '_cb_service_exit', '_cb_send_net_request', '_cb_get_key_id')

    def __init__(self, dist_logger: logging.Logger, service_id: int, service_full_name: str,
                 thread_args: Tuple[logging.Logger, Type[Service], Mapping, Callable[..., None], Callable[..., None]],
                 cb_get_key_id: Callable[[EnvironmentKey], int],
                 cb_send_net_request: Callable[..., None]):
        """Constructor."""
        self.logger_dist = dist_logger  # node distributed logger

        self.service_id = service_id  # service instance id
        self._cb_send_net_request = cb_send_net_request
        self._cb_get_key_id = cb_get_key_id

        self._service_calls = self._init_call_list()

        self._start_time = 0.0  # start timestamp of current run

        self._events: Set[Event] = set()  # set of current events (from any service call in any service thread)
        self._event_set_lock = Lock()  # thread safe access to self._events

        self._thread_event = Event()  # event controlling blocks in the main thread method

        self._is_running = False  # True when service is running (service_run() is invoked)
        self._do_exit = False  # True when service thread is asked to terminate

        self._cv = Condition()
        self._recv_queue: Deque[Tuple[int, bytes]] = collections.deque()  # incoming data from other services

        self._interrupted = False

        self.service_thread = Thread(
            target=self._thread_entry,
            args=thread_args,
            name=service_full_name,
            daemon=True)

    @property
    def name(self) -> str:
        """Returns the full service name."""
        return self.service_thread.name

    # start/stop/exit methods are called from the main thread
    def start(self) -> None:
        """Invoke the service_run() method.  Service must have instantiated (thread is active)."""
        if not self.is_active():
            raise ServiceControlError("service thread is not active")

        if self._do_exit:
            raise ServiceControlError("service thread is requested to terminate")

        if self._thread_event.is_set():
            raise ServiceControlError("service was already requested to start")

        if self._is_running:
            raise ServiceControlError("service is already running")

        self._thread_event.set()  # unblock service thread

    def stop(self) -> None:
        """Interrupt the service."""
        if not self.is_active():
            raise ServiceControlError("service thread is not active")

        if self._do_exit:
            raise ServiceControlError("service thread is requested to terminate")

        if self._interrupted:
            raise ServiceControlError("service was already requested to stop")

        if not self._is_running:
            raise ServiceControlError("service is already stopped")

        self._interrupt_service()

    def exit(self) -> None:
        """Shuts down the service thread."""
        if not self.is_active():
            raise ServiceControlError("service thread is not active")

        if self._do_exit:
            raise ServiceControlError("service thread was already requested to terminate")

        self._do_exit = True
        self._interrupt_service()
        self._thread_event.set()

    def is_active(self) -> bool:
        """Return if service thread is active (thread is running)."""
        return self.service_thread.is_alive()

    def is_running(self) -> bool:
        """Return if service is running (service_run() method invoked)"""
        return self._is_running

    def is_interrupted(self) -> bool:
        """Return if the service has been requested to stop."""
        return self._interrupted

    def _thread_entry(self, service_logger: logging.Logger, service_cls, service_config: Mapping,
                      cb_service_init: Callable[..., None],
                      cb_service_exit: Callable[..., None]):
        """Point of entry for a thread of execution."""
        try:
            service_instance: Service = service_cls(service_config['parameters'], service_logger,
                                                    lambda call_index, *args: self._service_calls[call_index](*args))
        except Exception as ex:
            # need to catch everything here - exception occurred in service init
            self.logger_dist.error("Instantiation failed due to exception: %s", ex)
            self._do_exit = True
            cb_service_init(self, ex, traceback.format_exc())
            return

        # instantiation successful
        cb_service_init(self)
        self.logger_dist.info("Instantiated")

        while not self._do_exit:
            self._thread_event.wait()

            if self._do_exit:
                break

            assert not len(self._events)

            self._service_calls = self._init_call_list()

            self._cb_send_net_request(NetProtocolID.HUB_SERVICE_STARTED, request_data=(self.service_id,))
            self.logger_dist.info("Started")

            # Order here is important for thread safety on start().  start() checks for event.is_set() first, so by
            # clearing the set flag second, we are guaranteed that is_running = True by the time a caller to start()
            # could get past the is_set() condition.
            self._is_running = True
            self._thread_event.clear()

            self._start_time = perf_counter()

            try:
                service_instance.service_run()
            except ServiceInterrupted:
                service_instance.service_exit()  # allow service to handle any cleanup tasks
            except Exception as ex:
                # need to catch everything here - exception occurred in service
                self.logger_dist.error("Terminating due to exception: %s", ex)

                self._do_exit = True
                # A start/stop/exit call could slip through, resulting in any of thread event being set and
                # service_interrupt() being called, both outcomes of which do not affect thread shutdown.
                self._is_running = False
                cb_service_exit(self, ex, traceback.format_exc())
                return

            self._cb_send_net_request(NetProtocolID.HUB_SERVICE_STOPPED, request_data=(self.service_id,))
            self.logger_dist.info("Stopped")

            # ordering here is important for thread safety of stop()
            self._is_running = False  # any start() call before we get to the wait is fine; wait() will return
            self._interrupted = False

        cb_service_exit(self)
        self.logger_dist.info("Exited")
        return

    def _init_call_list(self) -> List[Callable]:
        """Initialize the call list for a service run."""
        return [self._sc_interrupted,
                self._sc_sleep,
                self._sc_send_all,
                self._sc_send_to,
                self._sc_recv,
                self._sc_has_recv,
                self._sc_tell,
                self._sc_ask,
                self._sc_elapsed_time]

    def _interrupt_service(self) -> None:
        """Interrupt the service."""
        if self._interrupted:
            return

        self._interrupted = True

        # Replace all service calls except first with exception throw function.
        # First service call is to check for interrupted, so leave that one.
        # Service call replacement enables most service calls to not have to check for interrupted.
        self._service_calls[1:] = [no_call_interrupted] * (len(self._service_calls) - 1)

        with self._event_set_lock:
            for service_event in self._events:
                service_event.set()

        with self._cv:
            self._cv.notify_all()  # notify all threads associated with this service

    def _new_event(self) -> Event:
        """Create new Event instance."""
        new_event = Event()
        with self._event_set_lock:
            if self._interrupted:
                # set flag on this event so wait() will not block; do not add to set
                new_event.set()
                return new_event

            self._events.add(new_event)

        return new_event

    def _del_event(self, event: Event) -> None:
        """Remove the given Event."""
        with self._event_set_lock:
            # because these events are not reused, we don't need to clear() them
            self._events.remove(event)

    def recv_data(self, from_service_id: int, data: bytes) -> None:
        """Receive data from a peer, given by from_service_id.  Services can receive data even when not running."""
        with self._cv:
            self._recv_queue.append((from_service_id, data))
            self._cv.notify()  # service may or may not be blocked

    # service callbacks (service thread execution)
    def _sc_interrupted(self) -> bool:
        """Return interrupted flag."""
        return self._interrupted

    def _sc_elapsed_time(self) -> float:
        """Return the elapsed time from start of service run."""
        return perf_counter() - self._start_time

    def _sc_sleep(self, duration: Union[float, int]) -> None:
        """Suspend execution of service for the given amount of time (in seconds)."""
        self.logger_dist.debug("Sleeping for %.2f seconds ...", duration)
        service_event = self._new_event()
        service_event.wait(timeout=duration)
        self._del_event(service_event)

    def _sc_send_all(self, data: Any) -> None:
        """Send data to all services based on service class matching."""
        service_event = self._new_event()

        self._cb_send_net_request(NetProtocolID.HUB_SERVICE_SEND_ALL,
                                  request_data=(self.service_id, pickle.dumps(data)),
                                  local_data=service_event)
        service_event.wait()

        self._del_event(service_event)

    def _sc_send_to(self, to_service_ids: Tuple[int, ...], data: Any) -> None:
        """Send data to all services in list based on service class matching."""
        service_event = self._new_event()

        self._cb_send_net_request(NetProtocolID.HUB_SERVICE_SEND_TO,
                                  request_data=(self.service_id, pickle.dumps(to_service_ids), pickle.dumps(data)),
                                  local_data=service_event)

        service_event.wait()

        self._del_event(service_event)

    def _sc_has_recv(self) -> bool:
        """Returns True if recv queue is non-empty."""
        return len(self._recv_queue) > 0

    def _sc_recv(self) -> Tuple[int, Any]:
        """Wait until data is received through the Hub."""
        with self._cv:
            if self._interrupted:
                # It's possible that this method is entered on interrupt() but before the service call list is updated,
                # and the interrupt() method grabs the lock first.  In that case, when we enter here, without this
                # check, we may get stuck in a wait(), as notify_all() would have been called before we entered.
                # With the GIL, this wouldn't happen, but better to cover our bases.
                raise ServiceInterrupted

            while not len(self._recv_queue):
                # If a message is waiting already, simply return it, otherwise it is possible that
                # wait() may return without actually being notified, so check queue again...
                # A wait() call will release the lock, allowing another thread associated with this service to enter.
                # This may cause another wait(), or if the thread enters after service_recv_data() is called (it gets
                # the lock, then another service thread), it may grab the data, resulting in a notified thread staying
                # in the loop.  While the ordering may be indeterministic, this is thread safe for multiple service
                # threads to call.
                self._cv.wait()

                if self._interrupted:
                    # wait() call may have returned due to service interrupt request
                    raise ServiceInterrupted

            self.logger_dist.debug("Retrieving message (messages remaining in queue: %d)", len(self._recv_queue) - 1)
            from_service_id, data = self._recv_queue.popleft()
            return from_service_id, pickle.loads(data)

    def _sc_tell(self, obs_key: EnvironmentKey, obs_val: ENV_KEY_TYPE) -> None:
        """Give an observation to the network environment."""
        obs_key_id = self._cb_get_key_id(obs_key)

        s_obs = struct.pack(f">L{obs_key.key_list_type * len(obs_val)}", len(obs_val), *obs_val)

        service_event = self._new_event()

        self._cb_send_net_request(NetProtocolID.HUB_SERVICE_TELL,
                                  request_data=(self.service_id, obs_key_id, s_obs,),
                                  local_data=service_event)

        service_event.wait()

        self._del_event(service_event)

    def _sc_ask(self, ev_key: EnvironmentKey) -> ENV_KEY_TYPE:
        """Ask evidence from the network environment."""
        ev_key_id = self._cb_get_key_id(ev_key)

        ev_lst = []  # list to write evidence values to

        service_event = self._new_event()

        self._cb_send_net_request(NetProtocolID.HUB_SERVICE_ASK,
                                  request_data=(self.service_id, ev_key_id),
                                  local_data=(service_event, ev_key, ev_lst))
        service_event.wait()

        self._del_event(service_event)

        if self._interrupted:
            # wait() call may have returned due to service interrupt request
            raise ServiceInterrupted

        return tuple(ev_lst)



"""
eMews CORE network emulator service generator tool.
This tool supports CORE 9 and is compatible with CORE 8 for legacy purposes.
"""
import argparse
import pathlib

from ruamel.yaml import YAML

from emews.base.config import get_root_path

HELP_TXT = "CORE service generator - generates a CORE service interfacing the given eMews service by name"

parser = argparse.ArgumentParser(description='eMews CORE network emulator service generator tool')

parser.add_argument("emews_service_name",
                    help="Name of eMews service to be used to generate the respective CORE service")
parser.add_argument("-o", "--overwrite", action="store_true",
                    help="Overwrite existing CORE service of the same name")
parser.add_argument("--core8", action="store_true",
                    help="Generate CORE 8 compatible service - only use this if using CORE 8")

args = parser.parse_args()


def main() -> None:
    """Main function for service generator."""
    print(f"{HELP_TXT}\n")
    tool_service_gen()


def tool_service_gen() -> None:
    """CORE service generator tool."""
    emews_root = pathlib.Path(get_root_path())
    is_core8 = args.core8

    if is_core8:
        print("Generated service will be CORE 8 compatible")
        core_service_dir = 'core8_config_services'
    else:
        core_service_dir = 'core_services'

    core_service_path = emews_root.joinpath('core_emu', core_service_dir).resolve()

    if not core_service_path.exists():
        print("Could not find directory to write CORE service.  Expected location: {core_service_path}")
        return

    service_name: str = args.emews_service_name

    emews_service_path = emews_root.joinpath('services', service_name).resolve()

    if not emews_service_path.exists():
        print(f"Could not find eMews service: {service_name}.  Expected location: {emews_service_path}")
        return

    default_config_fullpath = emews_service_path.joinpath("default.yml").resolve()

    if not default_config_fullpath.exists():
        print(f"Could not find default service configuration.  Expected location: {default_config_fullpath}")
        return

    with open(default_config_fullpath) as f_config:
        config_dct = YAML().load(f_config)

    service_display_name = config_dct.get('display_name')

    if service_display_name is None:
        service_display_name = service_name

    if is_core8:
        core_service_template_name = "service_template_core8.py"
    else:
        core_service_template_name = "service_template.py"

    core_service_template = emews_root.joinpath('core_emu', core_service_template_name).read_text()
    core_service_fullpath = core_service_path.joinpath(f"{service_name}.py")

    print(f"- Using eMews service: {service_display_name}")
    print(f"- Writing CORE service to: {core_service_fullpath.resolve()}")

    is_overwrite = args.overwrite
    if is_overwrite:
        print("-- Will overwrite if CORE service already exists")
    print("\n[Generating CORE service] ...")

    service_exists = core_service_fullpath.exists()

    if service_exists and not is_overwrite:
        print(f"! {core_service_fullpath.name} not written - CORE service already exists (and overwrite disabled)")
        return

    additional_configs = ''
    additional_configs_txt = ''
    for s_file in emews_service_path.iterdir():
        if s_file.is_dir() or s_file.suffix != '.yml' and s_file.suffix != '.yaml' or s_file.name == 'default.yml':
            continue

        additional_configs = f'{additional_configs}", "{s_file.stem}'
        additional_configs_txt = f"{additional_configs_txt}\n   |- {s_file.name}"

    gen_service = core_service_template\
        .replace("__EMEWS_SERVICE_NAME__", service_name, 1)\
        .replace("__EMEWS_SERVICE_DISPLAY_NAME__", service_display_name, 1)\
        .replace("__EMEWS_CLASS_NAME__", service_name.capitalize(), 1)\
        .replace("__EMEWS_SERVICE_CONFIGS__", additional_configs, 1)

    with open(core_service_fullpath, "w") as f_service:
        f_service.write(gen_service)

    if service_exists:
        print(f"+ {core_service_fullpath.name} (existing service overwritten){additional_configs_txt}")
    else:
        print(f"+ {core_service_fullpath.name}{additional_configs_txt}")


main()



"""RNGs."""
import importlib
import random


class UniformInt:
    """Samples from a bounded discrete uniform distribution, [lower_bound, upper_bound]."""

    __slots__ = ('lower_bound', 'upper_bound')

    def __init__(self, lower_bound=0, upper_bound=1):
        """Constructor."""
        super().__init__()
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def sample(self):
        """Samples using a bounded uniform distribution."""
        return random.randint(self.lower_bound, self.upper_bound)

    def __str__(self) -> str:
        """Return a string of parameters and their current values."""
        return f"{self.__class__.__name__}: lower_bound={self.lower_bound}, " \
               f"upper_bound={self.upper_bound}"


class TruncnormInt:
    """
    Centered truncated discrete normal sampler.

    Mu is defined as the median in [lower_bound, upper_bound].
    """

    __slots__ = ('lower_bound', 'upper_bound', 'sigma', '_dist')

    def __init__(self, lower_bound=0, upper_bound=1, sigma=1):
        """Constructor."""
        super().__init__()

        # Import on object construction so scipy.stats isn't loaded if this sampler isn't needed.
        # Note: the memory overhead is substantial - 50mb + additional thread spawning.
        trunc_norm_obj = getattr(importlib.import_module('scipy.stats'), 'truncnorm')

        mu = upper_bound / 2.0
        self._dist = trunc_norm_obj((lower_bound - mu) / sigma, (upper_bound - mu) / sigma,
                                    loc=mu, scale=sigma)

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.sigma = sigma

    def sample(self):
        """Sample from a truncated normal distribution."""
        return int(round(self._dist.rvs(1)[0]))

    def __str__(self) -> str:
        """Return a string of parameters and their current values."""
        return f"{self.__class__.__name__}: lower_bound={self.lower_bound}, " \
               f"upper_bound={self.upper_bound}, sigma={self.sigma}"


class SequentialIterator:
    """
    Iterates through a range of numbers, returning the next number in a sequence.

    Once the last value in the sequence is sampled, that value is returned until the parameters are
    updated, at which point the sequence resets according to the new parameters.
    """

    __slots__ = ('end_value', '_current_value')

    def __init__(self, start_value=0, end_value=1):
        """Constructor."""
        super().__init__()

        self.end_value = end_value
        self._current_value = start_value

    def sample(self):
        """Return the next value sequentially from a list."""
        ret_value = self._current_value

        if self._current_value < self.end_value:
            self._current_value += 1

        return ret_value

    def __str__(self) -> str:
        """Return a string of parameters and their current values."""
        return f"{self.__class__.__name__}: current_value={self._current_value}, " \
               f"end_value={self.end_value}"



"""eMews metadata."""
# TODO: get this information from pyproject.toml

__version__ = "0.9.2"

__title__ = "eMews"
__summary__ = "Network traffic automation and prototyping."

__author__ = "Brian Ricks"
__email__ = "absolutefunk@utdallas.edu"

__license__ = "GNU General Public License, Version 3"
__copyright__ = "2017-2026 eMews Project"

_documentation_url = "https://mews.sv.cmu.edu/research/emews/"



"""
Base module for eMews network environments key structures.

This is a separate module to enable key importing without having to indirectly import env.  Only the hub needs to
import the environment module.
"""
from typing import Tuple, Union

from enum import Enum
import inspect
import os

ENV_KEY_BASE_TYPES = Union[bool, int, float]
ENV_KEY_TYPE = Tuple[ENV_KEY_BASE_TYPES, ...]

_ENV_KEY_IDX = 1  # free index for environment keys


class EnvironmentKey(Enum):
    """Environment observation and evidence key definitions."""

    def __new__(cls, key_list_type: ENV_KEY_BASE_TYPES):
        """Member value assigned, key types are passed in."""
        if key_list_type == bool:
            pack_type = '?'
        elif key_list_type == int:
            pack_type = 'l'
        elif key_list_type == float:
            pack_type = 'd'
        else:
            raise AttributeError(f"key list type '{key_list_type}' is not supported.  Use 'bool', 'int', or 'float'")

        global _ENV_KEY_IDX
        value = _ENV_KEY_IDX  # key id
        _ENV_KEY_IDX += 1

        # this is the module which contains the keys.py file
        _, key_parent_module, keys_module = inspect.getfile(cls).rsplit(os.sep, 2)

        if keys_module != 'keys.py':
            raise AttributeError("child EnvironmentKey class must be in a module named 'keys.py' "
                                 f"(given module name: {keys_module})")

        if cls.__name__ == 'Observation':
            key_type_abbrev = 'o'
        elif cls.__name__ == 'Evidence':
            key_type_abbrev = 'e'
        else:
            raise AttributeError("child EnvironmentKey class must be named 'Observation' or "
                                 f"'Evidence' (given class name: {cls.__name__})")

        obj = object.__new__(cls)
        obj._value_ = value
        obj.key_list_type = pack_type

        # this sets the full path name for the key, used with key name as a global identifier
        obj.key_env_path = f"{key_parent_module}.{key_type_abbrev}"
        return obj

    def key_full_name(self) -> str:
        """
        Return the full name of the key.

        This name is globally unique among all keys, and provides a means to lookup key ids.
        """
        return f"{self.key_env_path}.{self.name}"



"""DDoS Bots and Botnets."""
from typing import Tuple

import socket
import urllib3

from emews.api.service import Service, STR_ENCODING_UTF8


def init_udp_socket(packet_size) -> Tuple[socket.socket, bytes]:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    if packet_size < 9:
        payload = 'A'  # 1 byte payload (header is 8 bytes)
    else:
        payload = 'A' * (packet_size - 8)

    payload = bytes(payload, STR_ENCODING_UTF8)

    return sock, payload


class DefaultService(Service):
    """Simple UDP flooder.  No C&C."""

    __slots__ = ('_target_address', '_send_delay', '_packet_size', '_bytes_sent')

    def __init__(self, config):
        """Constructor."""
        super().__init__()

        self._target_address: str = config['target_address']  # IP address of target node to flood
        self._send_delay: float = config['send_delay']  # send pktsat this interval
        self._packet_size: int = config['packet_size']  # size of UDP pkt to send

        self._bytes_sent = 0

        self.logger.info("Params: target address: %s, packet send delay: %.2f, packet size: %d",
                         self._target_address, self._send_delay, self._packet_size)

    def service_run(self):
        self._bytes_sent = 0
        sock, payload = init_udp_socket(self._packet_size)

        target_address = self._target_address
        send_delay = self._send_delay

        while True:
            # upon interrupt, sleep() will raise ServiceInterrupted, invoking service_exit()
            self._bytes_sent += sock.sendto(payload, (target_address, 6200))  # just send to some port
            self.sleep(send_delay)

    def service_exit(self) -> None:
        self.logger.info("Total bytes sent: %d", self._bytes_sent)


class BotnetCC(Service):
    """Botnet C&C."""

    CMD_START = 1  # start flooding
    CMD_STOP = 2  # stop flooding

    __slots__ = ('_cc_id', '_flood_start_delay', '_flood_duration')

    def __init__(self, config):
        """Constructor."""
        super().__init__()

        self._flood_start_delay: int = config['flood_start_delay']  # delay (in seconds) to command bots to flood
        self._flood_duration: int = config['flood_duration']  # duration (in seconds) until bots commanded to stop

    def service_run(self) -> None:
        self.sleep(self._flood_start_delay)
        self.send_all(BotnetCC.CMD_START)
        self.sleep(self._flood_duration)
        self.send_all(BotnetCC.CMD_STOP)


class BotBase(Service):
    """Base class for bots."""

    __slots__ = ('_target_address', '_send_delay', '_sent_count')

    def __init__(self, config):
        """Constructor."""
        super().__init__()

        self._target_address: str = config['target_address']  # IP address of target node (to flood/induce a flood)
        self._send_delay: float = config['send_delay']  # send pkts/requests at this interval

        self._sent_count = 0  # number of requests or bytes sent

    def service_run(self) -> None:
        self._sent_count = 0

        cc_cmd = self.recv()[1]

        if cc_cmd == BotnetCC.CMD_START:
            # if bot starts after CC sends the start command, it will not flood
            self.logger.info("Commanded to start flood")

            send_delay = self._send_delay

            while not self.received_data():
                self._sent_count += self.bot_send()
                self.sleep(send_delay)

            if self.recv()[1] != BotnetCC.CMD_STOP:
                self.logger.warning("Received unknown command, expected 'CMD_STOP'")

        elif cc_cmd != BotnetCC.CMD_STOP:
            self.logger.warning("Received unknown command, expected 'CMD_START'")

        self.service_exit()

    def bot_send(self) -> int:
        """Send data to target."""
        raise NotImplementedError


class UDPFloodingBot(BotBase):
    """Bot-based UDP flooder."""

    __slots__ = ('_packet_size', '_sock', '_payload')

    def __init__(self, config):
        """Constructor."""
        super().__init__(config)

        self._packet_size: int = config['packet_size']  # size of UDP pkt to send
        self._sock, self._payload = init_udp_socket(self._packet_size)

        self.logger.info("Params: target address: %s, packet send delay: %.2f, packet size: %d",
                         self._target_address, self._send_delay, self._packet_size)

    def service_exit(self) -> None:
        self.logger.info("Total bytes sent: %d", self._sent_count)

    def bot_send(self) -> int:
        """Send UDP packets to target."""
        return self._sock.sendto(self._payload, (self._target_address, 6200))


class TCPFloodingBot(BotBase):
    """Bot-based TCP flooder, utilizing a request-based flood on the outbound from a target HTTPS server."""

    __slots__ = ('_http',)

    def __init__(self, config: dict):
        """Constructor."""
        super().__init__(config)

        # urllib3 related setup
        urllib3.disable_warnings()  # ignore self-signed SSL certs
        br_header = urllib3.util.make_headers(user_agent=config['user_agent'])
        self._http = urllib3.PoolManager(num_pools=2, headers=br_header, cert_reqs='CERT_NONE')

        self.logger.info("Params: target address: %s, request send delay: %.2f", self._target_address, self._send_delay)

    def service_exit(self) -> None:
        self.logger.info("Total requests sent: %d", self._sent_count)

    def bot_send(self) -> int:
        """Send TCP requests to target.  Induce outbound traffic."""
        try:
            self._http.request('GET', self._target_address)
        except Exception as ex:
            self.logger.warning("On request: %s, (target: %s)", ex, self._target_address)
            return 0

        return 1



"""Base module for eMews environments."""
from typing import Callable, Dict, List, Set, Tuple, Type

import importlib
import inspect
import logging
import os
import struct

from emews.base.config import get_environments_path
from emews.base.env_key import ENV_KEY_TYPE, EnvironmentKey
from emews.base.node import NetworkHandlerError, TimerEvent, EventCloseCB, TYPE_LEN


TYPE_LEN_INT = TYPE_LEN['L']


class NetEnvError(Exception):
    """Raised when a network environment or the manager encounters an issue."""
    pass


class Environment:
    """Network environment base class."""

    __slots__ = ('_logger', '_cb_env_call')

    def __init_subclass__(cls, **kwargs):
        """Subclass creation hook.  'cls' is the subclass (child Environment)."""
        super().__init_subclass__(**kwargs)
        child_init = cls.__init__

        def __init__(self: Environment, logger_dist: logging.Logger, cb_sys: Callable):
            """Child service __init__ override."""
            self._logger = logger_dist
            self._cb_env_call = cb_sys

            child_init(self)

        setattr(cls, '__init__', __init__)  # replace child class init with the def above

    @property
    def logger(self) -> logging.Logger:
        """System logger."""
        return self._logger

    def update_evidence(self, node_id: int, obs_key: EnvironmentKey, obs_val: ENV_KEY_TYPE) -> None:
        """Produce evidence, given the observation."""
        raise NotImplementedError

    def get_evidence(self, node_id: int, ev_key: EnvironmentKey) -> ENV_KEY_TYPE:
        """Return current evidence corresponding to the evidence key."""
        raise NotImplementedError

    def tell_to(self, obs_key: EnvironmentKey, obs_val: ENV_KEY_TYPE) -> None:
        """Give an observation to an environment."""
        self._cb_env_call(EnvState.CALL_INDEX_TELL_TO, obs_key, obs_val)

    def ask_from(self, ev_key: EnvironmentKey) -> ENV_KEY_TYPE:
        """Ask for evidence from an environment."""
        return self._cb_env_call(EnvState.CALL_INDEX_ASK_FROM, ev_key)

    def new_timer(self, interval: int, callback_func: Callable, repeat=True) -> int:
        """
        Create a new timer, in which callback_func is invoked every 'interval' seconds, unless repeat=False.

        Return a unique identifier for the timer.
        """
        return self._cb_env_call(EnvState.CALL_INDEX_NEW_TIMER, interval, callback_func, repeat=repeat)

    def del_timer(self, timer_id: int) -> None:
        """Remove the timer with identifier 'timer_id'."""
        self._cb_env_call(EnvState.CALL_INDEX_DEL_TIMER, timer_id)


class EnvState:
    """Environment state."""

    CALL_INDEX_TELL_TO = 0
    CALL_INDEX_ASK_FROM = 1
    CALL_INDEX_NEW_TIMER = 2
    CALL_INDEX_DEL_TIMER = 3

    __slots__ = ('logger_dist', 'keys_obs', 'keys_ev', 'env_update_evidence', 'env_get_evidence',
                 '_call_list', '_timer_events', '_timer_cb', '_cb_env_tell_to', '_cb_env_ask_from',
                 '_cb_env_timer', '_cb_event_close', 'module_name', '_env')

    def __init__(self, logger_dist: logging.Logger, module_name: str, env_cls: Type[Environment],
                 cb_env_tell_to: Callable, cb_env_ask_from: Callable, cb_env_timer: Callable,
                 cb_event_close: EventCloseCB):
        """Constructor."""
        self._call_list = [self._ec_tell_to,
                           self._ec_ask_from,
                           self._ec_new_timer,
                           self._ec_del_timer]

        self._timer_events: Dict[int, TimerEvent] = {}
        self._timer_cb: Dict[TimerEvent, Callable] = {}

        # env manager callbacks
        self._cb_env_tell_to = cb_env_tell_to
        self._cb_env_ask_from = cb_env_ask_from
        self._cb_env_timer = cb_env_timer
        self._cb_event_close = cb_event_close

        self.keys_obs: Set[EnvironmentKey] = set()
        self.keys_ev: Set[EnvironmentKey] = set()

        self.logger_dist = logger_dist
        self.module_name = module_name

        # ignore argument sig mismatch - args injected
        self._env = env_cls(self.logger_dist, lambda call_index, *args: self._call_list[call_index](*args))
        self.env_update_evidence = self._env.update_evidence
        self.env_get_evidence = self._env.get_evidence

    @property
    def full_name(self) -> str:
        return f"{self.module_name}.{self._env.__class__.__name__}"

    @property
    def obs_key_names(self) -> List[str]:
        """Return a listing of all observation environment keys by name."""
        return [obs_key.key_full_name() for obs_key in self.keys_obs]

    @property
    def ev_key_names(self) -> List[str]:
        """Return a listing of all evidence environment keys by name."""
        return [ev_key.key_full_name() for ev_key in self.keys_ev]

    def _handle_timer_expiry(self, timer_event: TimerEvent, num_expires: int):
        """Handle a timer expiry."""
        self._timer_cb[timer_event](timer_event.fileno(), num_expires)  # invoke env callback

    # environment callbacks
    def _ec_tell_to(self, obs_key: EnvironmentKey, obs_val: ENV_KEY_TYPE) -> None:
        """Give an observation to an environment."""
        self._cb_env_tell_to(self, obs_key, obs_val)

    def _ec_ask_from(self, ev_key: EnvironmentKey) -> ENV_KEY_TYPE:
        """Ask for evidence from an environment."""
        return self._cb_env_ask_from(self, ev_key)

    def _ec_new_timer(self, interval: int, callback_func: Callable, repeat) -> int:
        """
        Create a new timer, in which callback_func is invoked every 'interval' seconds, unless repeat=False.

        Return a unique identifier for the timer.
        """
        timer_event: TimerEvent = self._cb_env_timer(interval, self._handle_timer_expiry, repeat=repeat)
        timer_fd = timer_event.fileno()

        self._timer_cb[timer_event] = callback_func
        self._timer_events[timer_fd] = timer_event

        return timer_fd

    def _ec_del_timer(self, timer_id: int) -> None:
        """Remove the timer with identifier 'timer_id'."""
        try:
            timer_event = self._timer_events[timer_id]
        except KeyError:
            raise NetEnvError(f"environment timer {timer_id} does not exist")

        del self._timer_events[timer_id]
        del self._timer_cb[timer_event]
        self._cb_event_close(timer_event)


class EnvManager:
    """Manager for network environments."""

    __slots__ = ('logger_dist', '_cb_new_timer', '_env_state',
                 '_cb_event_close', '_key_id_by_name', '_key_by_id', '_env_state_by_key')

    def __init__(self, logger_dist: logging.Logger,
                 new_timer_cb: Callable, event_close_cb: EventCloseCB):
        """Constructor."""
        self.logger_dist = logger_dist
        self._cb_new_timer = new_timer_cb
        self._cb_event_close = event_close_cb

        self._key_id_by_name: Dict[str, int] = {}  # key name to its id (member value)
        self._key_by_id: Dict[int, EnvironmentKey] = {}  # key id to its member (enum)
        self._env_state_by_key: Dict[EnvironmentKey, EnvState] = {}  # env key to its env
        self._env_state: Set[EnvState] = set()

    @property
    def env_count(self) -> int:
        """Return the number of environments."""
        return len(self._env_state)

    @property
    def env_names(self) -> List[str]:
        """Return a listing of all environments by name."""
        return [env_state.full_name for env_state in self._env_state]

    @property
    def env_key_names(self) -> List[str]:
        """Return a listing of all environment keys by name."""
        key_lst = [key_obs.key_full_name() for env_state in self._env_state
                   for key_obs in env_state.keys_obs]
        key_lst.extend([key_ev.key_full_name() for env_state in self._env_state
                        for key_ev in env_state.keys_ev])
        return key_lst

    def parse_environments(self, logger: logging.Logger) -> None:
        """Instantiate all environments found."""
        with os.scandir(get_environments_path()) as it_environment_root:
            for entry_env_root in it_environment_root:
                env_module_name = None
                env_cls_obj = None
                keys_module_obj = None
                keys_import_fail = False

                if not entry_env_root.name.startswith('.') and entry_env_root.is_dir():
                    # may be a environment directory
                    env_module_name = entry_env_root.name

                    with os.scandir(entry_env_root.path) as it_dir:
                        # check if contents are consistent with an environment
                        for entry_dir in it_dir:
                            if entry_dir.is_file() and entry_dir.name == 'environment.py':
                                try:
                                    env_module_obj = importlib.import_module(
                                        f"emews.environments.{env_module_name}.environment")
                                except ImportError as ex:
                                    # apparently not a module
                                    logger.warning(
                                        "While parsing environments: directory '%s' contains "
                                        "environment.py but cannot be imported: %s",
                                        env_module_name, ex)
                                    break

                                env_cls_lst = []
                                # Get environment classes (only should be one to preserve 1:1
                                # mapping from keys - which environment to use is based on the
                                # env key itself, ie, environments are looked up based on key)
                                for name, obj in inspect.getmembers(env_module_obj):
                                    if inspect.isclass(obj) and issubclass(obj, Environment) and \
                                            obj != Environment:
                                        # environment class (child of Environment)
                                        env_cls_lst.append(obj)

                                if not len(env_cls_lst):
                                    logger.warning("While parsing environments: directory '%s' contains "
                                                   "environment.py but missing an Environment class",
                                                   env_module_name)
                                    break

                                if len(env_cls_lst) > 1:
                                    logger.warning("While parsing environments: directory '%s' contains "
                                                   "environment.py but more than one Environment class.",
                                                   env_module_name)
                                    break

                                env_cls_obj = env_cls_lst[0]

                            elif entry_dir.is_file() and entry_dir.name == 'keys.py':
                                try:
                                    keys_module_obj = importlib.import_module(
                                        f"emews.environments.{env_module_name}.keys")
                                except ImportError as ex:
                                    # apparently not a module (or EnvironmentKey creation failed)
                                    logger.warning("While parsing environments: directory '%s' contains "
                                                   "keys.py but cannot be imported: %s", env_module_name, ex)
                                    keys_import_fail = True
                                    break

                # Note, a directory missing an environment module but has a keys module will be
                # ignored (no warnings logged).
                if env_cls_obj is None or keys_import_fail:
                    # environment module or class missing, or keys module failed to import (this
                    # latter condition is here to prevent two warnings to log in terms of keys)
                    continue

                if keys_module_obj is None:
                    logger.warning("While parsing environments: directory '%s' contains "
                                   "environment.py but missing a keys module", env_module_name)
                    continue

                env_state = EnvState(self.logger_dist, env_module_name, env_cls_obj,
                                     self._env_tell_to, self._env_ask_from, self._cb_new_timer,
                                     self._cb_event_close)
                self._env_state.add(env_state)

                # get env keys (obs/ev)
                for cls_name, obj in inspect.getmembers(keys_module_obj):
                    if inspect.isclass(obj) and issubclass(obj, EnvironmentKey) and \
                            obj != EnvironmentKey:
                        if cls_name == "Observation":
                            env_keys = env_state.keys_obs
                        else:
                            # class naming scheme is enforced in EnvironmentKey construction
                            assert cls_name == "Evidence"
                            env_keys = env_state.keys_ev

                        for key_name, member in obj.__members__.items():
                            self._key_id_by_name[member.key_full_name()] = member.value
                            self._key_by_id[member.value] = member

                            env_keys.add(member)  # attribute of EnvState

                            self._env_state_by_key[member] = env_state

        if len(self._env_state):
            env_names = []
            for env_state in self._env_state:
                obs_names = env_state.obs_key_names
                if not len(obs_names):
                    obs_names.append("<none>")

                ev_names = env_state.ev_key_names
                if not len(ev_names):
                    ev_names.append("<none>")

                env_names.append(f"{env_state.full_name} [obs keys: {', '.join(obs_names)}, "
                                 f"ev keys: {', '.join(ev_names)}]")

            logger.info("%d environments are loaded: %s", len(env_names), ', '.join(env_names))
        else:
            logger.warning("No environments available to load")

    def _env_tell_to(self, from_env: EnvState, obs_key: EnvironmentKey, obs_val: ENV_KEY_TYPE) -> None:
        """Give an observation to environment according to observation key."""
        self.logger_dist.debug("New observation (env -> env) from environment '%s': key %s, value: %s",
                               from_env.full_name, obs_key.key_full_name(), str(obs_val))

        try:
            env_state = self._env_state_by_key[obs_key]
        except KeyError:
            raise NetEnvError(f"environment key '{obs_key.key_full_name()}' does not have an "
                              "environment associated to it")

        if obs_key not in env_state.keys_obs:
            raise NetEnvError(f"environment key '{obs_key.key_full_name()}' is not a valid "
                              f"observation key for environment {env_state.full_name} "
                              f"(observation from {from_env.full_name})")

        env_state.env_update_evidence(0, obs_key, obs_val)

    def _env_ask_from(self, from_env: EnvState, ev_key: EnvironmentKey) -> ENV_KEY_TYPE:
        """Ask for evidence from environment according to evidence key."""
        self.logger_dist.debug("Evidence request (env -> env) from environment '%s': key: %s",
                               from_env.full_name, ev_key.key_full_name())

        try:
            env_state = self._env_state_by_key[ev_key]
        except KeyError:
            raise NetEnvError(f"environment key '{ev_key.key_full_name()}' does not have an "
                              "environment associated to it")

        if ev_key not in env_state.keys_ev:
            raise NetEnvError(f"environment key '{ev_key.key_full_name()}' is not a valid evidence "
                              f"key for environment {env_state.full_name} (evidence request from "
                              f"{from_env.full_name})")

        return env_state.env_get_evidence(0, ev_key)

    # network handlers
    def env_handle_envkey_id_req(self, node_id: int, env_key_name_bytes: bytes) -> Tuple[int]:
        """Return environment key ID corresponding to given env key name."""
        env_key_name = env_key_name_bytes.decode()

        try:
            env_key_id = self._key_id_by_name[env_key_name]
        except KeyError:
            raise NetworkHandlerError(f"environment key '{env_key_name}' does not exist")

        self.logger_dist.debug("Node %d requested id for environment key '%s'",
                               node_id, env_key_name)

        return env_key_id,

    def env_handle_new_obs(self, node_id: int, service_id: int, obs_key_id: int, obs_val_packed: bytes) -> tuple:
        """Given an observation, update the environments to which the keys correspond (Tell)."""
        try:
            obs_key = self._key_by_id[obs_key_id]
        except KeyError:
            raise NetworkHandlerError(f"environment observation key id '{obs_key_id}' not found "
                                      f"(node id: {node_id}, service id: {service_id})")

        num_obs: int = struct.unpack_from(">L", obs_val_packed)[0]
        obs_val: Tuple = struct.unpack_from(f">{obs_key.key_list_type * num_obs}", obs_val_packed, offset=TYPE_LEN_INT)

        self.logger_dist.debug("New observation from node %d, service instance %d: key: %s, value: %s",
                               node_id, service_id, obs_key.key_full_name(), str(obs_val))

        env_state = self._env_state_by_key[obs_key]  # key will exist as an id mapping exists for it

        if obs_key not in env_state.keys_obs:
            raise NetworkHandlerError(f"environment key '{obs_key.key_full_name()}' is not a valid "
                                      f"observation key for environment {env_state.full_name}")

        try:
            env_state.env_update_evidence(node_id, obs_key, obs_val)
        except NetEnvError as ex:
            raise NetworkHandlerError(str(ex))

        return ()

    def env_handle_ev_request(self, node_id: int, service_id: int, ev_key_id: int) -> Tuple[bytes]:
        """Given an evidence key, fetch evidence (Ask)."""
        try:
            ev_key = self._key_by_id[ev_key_id]
        except KeyError:
            raise NetworkHandlerError(f"environment evidence key id '{ev_key_id}' not found "
                                      f"(node id: {node_id}, service id: {service_id})")

        self.logger_dist.debug("New evidence request from node %d, service instance %d: key: %s",
                               node_id, service_id, ev_key.key_full_name())

        env_state = self._env_state_by_key[ev_key]  # key will exist as an id mapping exists for it

        if ev_key not in env_state.keys_ev:
            raise NetworkHandlerError(f"environment key '{ev_key.key_full_name()}' is not a valid "
                                      f"evidence key for environment {env_state.full_name}")

        try:
            ev_val = env_state.env_get_evidence(node_id, ev_key)
        except NetEnvError as ex:
            raise NetworkHandlerError(str(ex)) from ex

        return struct.pack(f">L{ev_key.key_list_type * len(ev_val)}", len(ev_val), *ev_val),



"""eMews server CORE Services."""
from typing import Any

from core.config import Configuration
from core.nodes.interface import CoreInterface
from core.services.base import CoreService, ServiceMode, ShadowDir

GROUP_NAME: str = "eMews"

class HttpsService(CoreService):
    """Apache HTTP/HTTPS Server."""

    name: str = "HTTPS"
    group: str = GROUP_NAME
    directories: list[str] = [
        "/etc/apache2",
        "/var/run/apache2",
        "/var/log/apache2",
        "/run/lock",
        "/var/lock/apache2",
        "/var/run/apache2/ssl",
        "/var/run/apache2/ssl/cert",
        "/var/run/apache2/ssl/key",
    ]
    files: list[str] = [
        "/etc/apache2/apache2.conf",
        "/etc/apache2/envvars",
        "apache2_start.sh"
    ]
    executables: list[str] = ["apache2ctl", "openssl"]
    dependencies: list[str] = []
    startup: list[str] = ["chown www-data /var/lock/apache2", "sh apache2_start.sh"]
    validate: list[str] = ["pidof apache2"]
    shutdown: list[str] = ["apache2ctl stop"]
    validation_mode: ServiceMode = ServiceMode.BLOCKING
    default_configs: list[Configuration] = []
    modes: dict[str, dict[str, str]] = {}
    # mount HTML root from file system (this is the default path for Apache2 on Ubuntu)
    shadow_directories: list[ShadowDir] = [ShadowDir(path="/var/www/html")]

    def data(self) -> dict[str, Any]:
        ifaces: list[CoreInterface] = []

        for iface in self.node.get_ifaces(control=False):
            ifaces.append(iface)
        return dict(ifaces=ifaces)



"""Environment keys for web environment."""
from emews.api.env_key import EnvironmentKey


class Observation(EnvironmentKey):
    """Observation environment key enums.  Observation key class must be named 'Observation'."""
    CRAWL_SITE = int
    LINK_CLICKED = int


class Evidence(EnvironmentKey):
    """Evidence environment key enums.  Evidence key class must be named 'Evidence'."""
    VIRAL_LINKS = int



"""HTTPS legacy CORE service.  Adapted to CORE 8 from CORE 5 code."""
from core import utils
from core.errors import CoreCommandError
from core.nodes.base import CoreNode
from core.services.coreservices import CoreService


class HttpsService(CoreService):
    """Start an apache server (HTTPS support)."""

    name = "Serv_HTTPS"
    group = "eMews"
    configs = ("/etc/apache2/apache2.conf", "/etc/apache2/envvars", "httpd_start.sh",)
    dirs = ("/etc/apache2", "/var/run/apache2", "/var/log/apache2",
            "/run/lock", "/var/lock/apache2", "/var/run/apache2/ssl",
            "/var/run/apache2/ssl/cert", "/var/run/apache2/ssl/key",)
    startup = ("chown www-data /var/lock/apache2", "sh httpd_start.sh",)
    startindex = 50
    shutdown = ("apache2ctl stop",)
    validate = ("pidof apache2",)

    APACHEVER22, APACHEVER24 = (22, 24)

    @classmethod
    def generate_config(cls, node: CoreNode, filename):
        """Generate an apache2.conf configuration file."""
        if filename == cls.configs[0]:
            return cls.generateapache2conf(node, filename)
        elif filename == cls.configs[1]:
            return cls.generateenvvars(node, filename)
        elif filename == cls.configs[2]:
            return cls.generatekeygen(node, filename)
        else:
            return ""

    @classmethod
    def detectversionfromcmd(cls) -> int:
        """
        Detect the apache2 version using the 'a2query' command.
        """
        try:
            result = utils.cmd("a2query -v")
            status = 0
        except CoreCommandError as e:
            status = e.returncode
            result = e.stderr
        if status == 0 and result[:3] == "2.4":
            return cls.APACHEVER24
        return cls.APACHEVER22

    @classmethod
    def generateapache2conf(cls, node, filename):
        lockstr = {cls.APACHEVER22:
                       'LockFile ${APACHE_LOCK_DIR}/accept.lock\n',
                   cls.APACHEVER24:
                       'Mutex file:${APACHE_LOCK_DIR} default\n', }
        mpmstr = {cls.APACHEVER22: '', cls.APACHEVER24:
            'LoadModule mpm_worker_module /usr/lib/apache2/modules/mod_mpm_worker.so\n', }

        permstr = {cls.APACHEVER22:
                       '    Order allow,deny\n    Deny from all\n    Satisfy all\n',
                   cls.APACHEVER24:
                       '    Require all denied\n', }

        authstr = {cls.APACHEVER22:
                       'LoadModule authz_default_module /usr/lib/apache2/modules/mod_authz_default.so\n',
                   cls.APACHEVER24:
                       'LoadModule authz_core_module /usr/lib/apache2/modules/mod_authz_core.so\n', }

        permstr2 = {cls.APACHEVER22:
                        '\t\tOrder allow,deny\n\t\tallow from all\n',
                    cls.APACHEVER24:
                        '\t\tRequire all granted\n', }

        version = cls.detectversionfromcmd()
        cfg = "# apache2.conf generated by utility.py:HttpService\n"
        cfg += lockstr[version]
        cfg += """\
PidFile ${APACHE_PID_FILE}
Timeout 300
KeepAlive On
MaxKeepAliveRequests 100
KeepAliveTimeout 5
"""
        cfg += mpmstr[version]
        cfg += """\

<IfModule mpm_prefork_module>
    StartServers          5
    MinSpareServers       5
    MaxSpareServers      10
    MaxClients          150
    MaxRequestsPerChild   0
</IfModule>

<IfModule mpm_worker_module>
    StartServers          2
    MinSpareThreads      25
    MaxSpareThreads      75
    ThreadLimit          64
    ThreadsPerChild      25
    MaxClients          150
    MaxRequestsPerChild   0
</IfModule>

<IfModule mpm_event_module>
    StartServers          2
    MinSpareThreads      25
    MaxSpareThreads      75
    ThreadLimit          64
    ThreadsPerChild      25
    MaxClients          150
    MaxRequestsPerChild   0
</IfModule>

User ${APACHE_RUN_USER}
Group ${APACHE_RUN_GROUP}

AccessFileName .htaccess

<Files ~ "^\.ht">
"""
        cfg += permstr[version]
        cfg += """\
</Files>

DefaultType None

HostnameLookups Off

ErrorLog ${APACHE_LOG_DIR}/error.log
LogLevel warn

#Include mods-enabled/*.load
#Include mods-enabled/*.conf
LoadModule alias_module /usr/lib/apache2/modules/mod_alias.so
LoadModule auth_basic_module /usr/lib/apache2/modules/mod_auth_basic.so
"""
        cfg += authstr[version]
        cfg += """\
LoadModule authz_host_module /usr/lib/apache2/modules/mod_authz_host.so
LoadModule authz_user_module /usr/lib/apache2/modules/mod_authz_user.so
LoadModule autoindex_module /usr/lib/apache2/modules/mod_autoindex.so
LoadModule dir_module /usr/lib/apache2/modules/mod_dir.so
LoadModule env_module /usr/lib/apache2/modules/mod_env.so
LoadModule ssl_module /usr/lib/apache2/modules/mod_ssl.so

NameVirtualHost *:80
Listen 80
Listen 443

LogFormat "%v:%p %h %l %u %t \\"%r\\" %>s %O \\"%{Referer}i\\" \\"%{User-Agent}i\\"" vhost_combined
LogFormat "%h %l %u %t \\"%r\\" %>s %O \\"%{Referer}i\\" \\"%{User-Agent}i\\"" combined
LogFormat "%h %l %u %t \\"%r\\" %>s %O" common
LogFormat "%{Referer}i -> %U" referer
LogFormat "%{User-agent}i" agent

ServerTokens OS
ServerSignature On
TraceEnable Off

<VirtualHost *:80>
    ServerAdmin webmaster@cmu.sv.mews
    DocumentRoot /var/www/html
    <Directory />
        Options FollowSymLinks
        AllowOverride None
    </Directory>
    <Directory /var/www/html/>
        Options Indexes FollowSymLinks MultiViews
        AllowOverride None
"""
        cfg += permstr2[version]
        cfg += """\
    </Directory>
    ErrorLog ${APACHE_LOG_DIR}/error.log
    LogLevel warn
    CustomLog ${APACHE_LOG_DIR}/access.log combined
</VirtualHost>
<VirtualHost *:443>
    SSLEngine on
    SSLVerifyClient none
    SSLCertificateFile "/var/run/apache2/ssl/cert/httpd.crt"
    SSLCertificateKeyFile "/var/run/apache2/ssl/key/httpd.key"
    ServerAdmin webmaster@cmu.sv.mews
    DocumentRoot /var/www/html
    <Directory />
        Options FollowSymLinks
        AllowOverride None
    </Directory>
    <Directory /var/www/html/>
        Options Indexes FollowSymLinks MultiViews
        AllowOverride None
"""
        cfg += permstr2[version]
        cfg += """\
        </Directory>
        ErrorLog ${APACHE_LOG_DIR}/error.log
        LogLevel warn
        CustomLog ${APACHE_LOG_DIR}/access.log combined
</VirtualHost>
"""


        return cfg

    @classmethod
    def generateenvvars(cls, node, filename):
        return """\
# this file is used by apache2ctl - generated by utility.py:HttpService
# these settings come from a default Ubuntu apache2 installation
export APACHE_RUN_USER=www-data
export APACHE_RUN_GROUP=www-data
export APACHE_PID_FILE=/var/run/apache2.pid
export APACHE_RUN_DIR=/var/run/apache2
export APACHE_LOCK_DIR=/var/lock/apache2
export APACHE_LOG_DIR=/var/log/apache2
export LANG=C
export LANG
"""

    @classmethod
    def generatekeygen(cls, node: CoreNode, filename):
        """Generate a cert for Apache to use."""
        cert_cn = node.name
        for iface in node.get_ifaces():
            if iface.name == "eth0":
                cert_cn = str(iface.get_ip4()).split("/")[0]  # m4d h4x
                break
        cfg = """\
#!/bin/sh
# We assume that the node only has one interface, so we assign to the cert the IP of this interface.
# Key size of 1024 is deprecated
openssl req -new -x509 -days 365 -sha1 -newkey rsa:2048 -nodes -keyout /var/run/apache2/ssl/key/httpd.key -out /var/run/apache2/ssl/cert/httpd.crt -subj '/O=CORE/OU=CORE-emu/CN="""
        cfg += cert_cn
        cfg += """'
# Launch Apache here so we know that cert/key was generated first.
apache2ctl start
"""
        return cfg



"""Automates the process of a user interacting with an SSH client."""
from typing import List

from pexpect import pxssh

from emews.api.random import TruncnormInt, UniformInt
from emews.api.service import Service


class DefaultService(Service):
    """Classdocs."""

    __slots__ = ('_ssh_session', '_host', '_port', '_username', '_password', '_command_list',
                 '_num_commands_sampler', '_command_sampler_sigma', '_command_delay_sampler',
                 '_crawl_sampler')

    def __init__(self, config: dict):
        """Constructor."""
        super().__init__()

        self._host = config['host']
        self._port = config['port']
        self._username = config['username']
        self._password = config['password']

        self._command_list: List[str] = config['command_list']

        self._num_commands_sampler = TruncnormInt(**config['num_commands_sampler'])
        self._command_delay_sampler = UniformInt(**config['command_delay_sampler'])
        self._crawl_sampler = UniformInt(**config['crawl_sampler'])

        self._command_sampler_sigma = config['command_sampler']['sigma']

        self._ssh_session = None

    def service_run(self):
        """Connect and login to the ssh server given with the credentials given."""
        self.logger.debug("Distributions: num_commands: %s, command_delay: %s, crawl: %s, "
                          "command: sigma=%s (upper_bound set dynamically)",
                          str(self._num_commands_sampler), str(self._command_delay_sampler),
                          str(self._command_delay_sampler), str(self._command_sampler_sigma))

        while not self.interrupted():
            self.sleep(self._crawl_sampler.sample())
            self._ssh_crawl()

    def service_exit(self) -> None:
        """Close the connection upon service interrupted while blocked."""
        ssh_client = self._ssh_session
        if ssh_client is not None:
            ssh_client.close()

    def _ssh_crawl(self):
        """
        Perform a single login and SSH crawl.

        Note, if using CORE, a new key-pair is generated for each network session.  This can result in SSH connections
        failing after one session, due to a new key-pair which doesn't match with what is in /root/.ssh/known_hosts.
        Assuming CORE is being run on a dedicated system, delete the known_hosts file before every CORE session.
        """
        ssh_client = pxssh.pxssh()
        self._ssh_session = ssh_client

        self.logger.info("Connecting to SSH server at %s:%d, user='%s' ...",
                         self._host, self._port, self._username)

        ssh_client.force_password = True
        ssh_client.login(self._host,
                         self._username,
                         password=self._password,
                         port=self._port)

        self.logger.info("Connected to SSH server (%s@%s:%d), executing commands ...", self._username, self._host,
                         self._port)

        # As we are sampling without replacement, we need to copy the original list
        command_list = list(self._command_list)
        cmd_sigma = self._command_sampler_sigma

        executed_commands: List[str] = []

        # loop until command count reached
        num_commands = self._num_commands_sampler.sample()
        for _ in range(num_commands):
            command_sampler = TruncnormInt(upper_bound=len(command_list) - 1, sigma=cmd_sigma)
            next_command = command_list.pop(command_sampler.sample())

            executed_commands.append(next_command)

            ssh_client.sendline(next_command)
            ssh_client.prompt()

            self.sleep(self._command_delay_sampler.sample())

        self.logger.info("Commands executed: [%s]", ", ".join(executed_commands))
        self.logger.info("Logging out (%s@%s:%d) ...", self._username, self._host, self._port)

        ssh_client.logout()
        ssh_client.close()
        self._ssh_session = None
